"""Routes for Tasks."""
from http import HTTPStatus as StatusCodes
import json
from typing import Any, cast, Dict, List, Set, Union

from flask import jsonify, request
from sqlalchemy import desc, insert, select, tuple_, update
from sqlalchemy.dialects.mysql import insert as mysql_insert
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.exc import DataError, IntegrityError
from sqlalchemy.orm import Session
from sqlalchemy.sql import func
import structlog

from jobmon.core import constants
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.task_arg import TaskArg
from jobmon.server.web.models.task_attribute import TaskAttribute
from jobmon.server.web.models.task_attribute_type import TaskAttributeType
from jobmon.server.web.models.task_instance import TaskInstance
from jobmon.server.web.models.task_instance_error_log import TaskInstanceErrorLog
from jobmon.server.web.models.task_resources import TaskResources
from jobmon.server.web.models.task_status import TaskStatus
from jobmon.server.web.models.workflow import Workflow
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint
from jobmon.server.web.server_side_exception import InvalidUsage, ServerError


logger = structlog.get_logger(__name__)


@blueprint.route("/task/bind_tasks_no_args", methods=["PUT"])
def bind_tasks_no_args() -> Any:
    """Bind the task objects to the database."""
    all_data = cast(Dict, request.get_json())
    tasks = all_data["tasks"]
    workflow_id = int(all_data["workflow_id"])
    mark_created = bool(all_data["mark_created"])
    structlog.contextvars.bind_contextvars(workflow_id=workflow_id)
    logger.info("Binding tasks")
    # receive from client the tasks in a format of:
    # {<hash>:[node_id(1), task_args_hash(2), array_id(3), task_resources_id(4), name(5),
    # command(6), max_attempts(7), reset_if_running(8),resource_scales(9),
    # fallback_queues(10) }

    session = SessionLocal()
    with session.begin():
        # Retrieve existing task_ids
        task_select_stmt = select(Task).where(
            (Task.workflow_id == workflow_id),
            tuple_(Task.node_id, Task.task_args_hash).in_(
                [tuple_(task[0], task[1]) for task in tasks.values()]
            ),
        )
        prebound_tasks = session.execute(task_select_stmt).scalars().all()

        # Bind tasks not present in DB
        tasks_to_add: List[
            Dict
        ] = []  # Container for tasks not yet bound to the database
        present_tasks = {
            (task.node_id, task.task_args_hash): task for task in prebound_tasks
        }  # Dictionary mapping existing Tasks to the supplied arguments
        # Dict mapping input tasks to the corresponding args/attributes
        task_hash_lookup = {}  # Reverse dictionary of inputs, maps hash back to values
        for hashval, items in tasks.items():
            (
                node_id,
                arg_hash,
                array_id,
                task_resources_id,
                name,
                command,
                max_att,
                reset,
                resource_scales,
                fallback_queues,
            ) = items

            id_tuple = (node_id, arg_hash)

            # Conditional logic: Has task already been bound to the DB? If yes, reset the
            # task status and update the args/attributes
            if id_tuple in present_tasks.keys():
                task = present_tasks[id_tuple]
                task.reset(
                    name=name,
                    command=command,
                    max_attempts=max_att,
                    reset_if_running=reset,
                )

            # If not, add the task
            else:
                task = {
                    "workflow_id": workflow_id,
                    "node_id": node_id,
                    "task_args_hash": arg_hash,
                    "array_id": array_id,
                    "task_resources_id": task_resources_id,
                    "name": name,
                    "command": command,
                    "max_attempts": max_att,
                    "status": constants.TaskStatus.REGISTERING,
                    "resource_scales": str(resource_scales),
                    "fallback_queues": str(fallback_queues),
                }
                tasks_to_add.append(task)

            task_hash_lookup[id_tuple] = hashval

        # Update existing tasks
        if present_tasks:
            # ORM task objects already updated in task.reset, flush the changes
            session.flush()

        # Bind new tasks with raw SQL
        if len(tasks_to_add):
            # This command is guaranteed to succeed, since names are truncated in the client
            task_insert_stmt = insert(Task).values(tasks_to_add)
            session.execute(task_insert_stmt)
            session.flush()

            # Fetch newly bound task ids
            new_task_query = select(Task).where(
                (Task.workflow_id == workflow_id),
                tuple_(Task.node_id, Task.task_args_hash).in_(
                    [
                        tuple_(task["node_id"], task["task_args_hash"])
                        for task in tasks_to_add
                    ]
                ),
            )
            new_tasks = session.execute(new_task_query).scalars().all()

        else:
            # Empty task list
            new_tasks = []

        # Create the response dict of tasks {<hash>: [id, status]}
        # Done here to prevent modifying tasks, and necessitating a refresh.
        return_tasks = {}

        for task in prebound_tasks + new_tasks:
            id_tuple = (task.node_id, task.task_args_hash)
            hashval = task_hash_lookup[id_tuple]
            return_tasks[hashval] = [task.id, task.status]

        # Set the workflow's created date if this is the last chunk of tasks.
        # Mark that a workflow has completed binding
        if mark_created:
            session.execute(
                update(Workflow)
                .where(Workflow.id == workflow_id, Workflow.created_date.is_(None))
                .values(created_date=func.now())
            )

    resp = jsonify(tasks=return_tasks)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task/bind_task_args", methods=["PUT"])
def bind_task_args() -> Any:
    """Add task args and associated task ids to the database."""
    all_data = cast(Dict, request.get_json())
    task_args = all_data["task_args"]
    if any(task_args):
        # Insert task args using INSERT IGNORE to handle conflicts
        task_arg_values = [
            {"task_id": task_id, "arg_id": arg_id, "val": value}
            for task_id, arg_id, value in task_args
        ]
        session = SessionLocal()
        try:
            if SessionLocal.bind.dialect.name == "mysql":
                arg_insert_stmt = (
                    insert(TaskArg).values(task_arg_values).prefix_with("IGNORE")
                )
            elif SessionLocal.bind.dialect.name == "sqlite":
                arg_insert_stmt = (
                    sqlite_insert(TaskArg)
                    .values(task_arg_values)
                    .on_conflict_do_nothing()
                )
            else:
                raise ServerError(
                    "invalid sql dialect. Only (mysql, sqlite) are supported. Got"
                    + SessionLocal.bind.dialect.name
                )
            with session.begin():
                session.execute(arg_insert_stmt)
        except (DataError, IntegrityError) as e:
            # Args likely too long, message back
            raise InvalidUsage(
                "Task Args are constrained to 1000 characters, you may have values "
                f"that are too long. Message: {str(e)}",
                status_code=400,
            ) from e
    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task/bind_task_attributes", methods=["PUT"])
def bind_task_attributes() -> Any:
    """Add task attributes and associated attribute types to the database."""
    all_data = cast(Dict, request.get_json())
    attributes = all_data["task_attributes"]

    # Map attribute names to attribute_type_ids, insert if necessary
    all_attribute_names = set()
    for attribute in attributes.values():
        all_attribute_names |= set(attribute.keys())

    if any(all_attribute_names):
        session = SessionLocal()
        with session.begin():
            attribute_type_ids = _add_or_get_attribute_types(
                all_attribute_names, session
            )
            # Build our insert values. On conflicts, update the existing value
            insert_values = []
            for task_id, attribute_dict in attributes.items():
                for attribute_name, attribute_val in attribute_dict.items():
                    insert_row = {
                        "task_id": task_id,
                        "task_attribute_type_id": attribute_type_ids[attribute_name],
                        "value": attribute_val,
                    }
                    insert_values.append(insert_row)

            # Insert and handle the conflicts
            if insert_values:
                try:
                    if SessionLocal.bind.dialect.name == "mysql":
                        attr_insert_stmt = mysql_insert(TaskAttribute).values(
                            insert_values
                        )
                        attr_insert_stmt = attr_insert_stmt.on_duplicate_key_update(
                            value=attr_insert_stmt.inserted.value
                        )
                        session.execute(attr_insert_stmt)

                    elif SessionLocal.bind.dialect.name == "sqlite":
                        for attr_to_add in insert_values:
                            attr_insert_stmt = (
                                sqlite_insert(TaskAttribute)
                                .values(attr_to_add)
                                .on_conflict_do_update(
                                    index_elements=[
                                        "task_id",
                                        "task_attribute_type_id",
                                    ],
                                    set_=dict(value=attr_to_add["value"]),
                                )
                            )
                            session.execute(attr_insert_stmt)
                    else:
                        raise ServerError(
                            "invalid sql dialect. Only (mysql, sqlite) are supported. Got"
                            + SessionLocal.bind.dialect.name
                        )
                except (DataError, IntegrityError) as e:
                    # Attributes too long, message back
                    raise InvalidUsage(
                        "Task attributes are constrained to 255 characters, "
                        f"you may have values that are too long. Message: {str(e)}",
                        status_code=400,
                    ) from e

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp


def _add_or_get_attribute_types(
    names: Union[List[str], Set[str]], session: Session
) -> Dict[str, int]:
    # Query for existing attribute types, to avoid integrity conflicts
    names = set(names)

    existing_rows_select = select(TaskAttributeType).where(
        TaskAttributeType.name.in_(names)
    )
    existing_rows = session.execute(existing_rows_select).scalars()
    existing_rows = {attr.name: attr.id for attr in existing_rows}

    existing_names = set(existing_rows.keys())

    # Insert the remaining names, found from the difference between old and new
    # Keep the IGNORE prefix in case other agents add attributes first, prevent errors
    # while trying to minimize collisions
    new_names = names - existing_names

    # We'll eventually return the combination of existing + new attribute type ids
    return_dict = existing_rows
    if any(new_names):
        new_attribute_types = [{"name": name} for name in new_names]
        try:
            if SessionLocal.bind.dialect.name == "mysql":
                insert_stmt = (
                    insert(TaskAttributeType)
                    .values(new_attribute_types)
                    .prefix_with("IGNORE")
                )
            elif SessionLocal.bind.dialect.name == "sqlite":
                insert_stmt = (
                    sqlite_insert(TaskAttributeType)
                    .values(new_attribute_types)
                    .on_conflict_do_nothing()
                )
            else:
                raise ServerError(
                    "invalid sql dialect. Only (mysql, sqlite) are supported. Got"
                    + SessionLocal.bind.dialect.name
                )
            session.execute(insert_stmt)

        except DataError as e:
            raise InvalidUsage(
                "Attribute types are constrained to 255 characters, your "
                f"attributes might be too long. Message: {str(e)}",
                status_code=400,
            ) from e

        # Query the IDs of the newly inserted rows
        new_rows_select = select(TaskAttributeType).where(
            TaskAttributeType.name.in_(new_names)
        )
        new_attribute_type_ids = session.execute(new_rows_select).scalars().all()

        # Update our return dict
        return_dict.update(
            {attribute.name: attribute.id for attribute in new_attribute_type_ids}
        )
    return return_dict


@blueprint.route("/task/bind_resources", methods=["POST"])
def bind_task_resources() -> Any:
    """Add the task resources for a given task."""
    data = cast(Dict, request.get_json())

    session = SessionLocal()
    with session.begin():
        new_resources = TaskResources(
            queue_id=data["queue_id"],
            task_resources_type_id=data.get("task_resources_type_id", None),
            requested_resources=json.dumps(data.get("requested_resources", None)),
        )
        session.add(new_resources)

    resp = jsonify(new_resources.id)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task/<task_id>/most_recent_ti_error", methods=["GET"])
def get_most_recent_ti_error(task_id: int) -> Any:
    """Route to determine the cause of the most recent task_instance's error.

    Args:
        task_id (int): the ID of the task.

    Return:
        error message
    """
    structlog.contextvars.bind_contextvars(task_id=task_id)
    logger.info(f"Getting most recent ji error for ti {task_id}")

    session = SessionLocal()
    with session.begin():
        select_stmt = (
            select(TaskInstanceErrorLog)
            .join_from(
                TaskInstance,
                TaskInstanceErrorLog,
                TaskInstance.id == TaskInstanceErrorLog.task_instance_id,
            )
            .where(TaskInstance.task_id == task_id)
            .order_by(desc(TaskInstance.id))
            .limit(1)
        )
        ti_error = session.execute(select_stmt).scalars().one_or_none()

    if ti_error is not None:
        resp = jsonify(
            {
                "error_description": ti_error.description,
                "task_instance_id": ti_error.task_instance_id,
            }
        )
    else:
        resp = jsonify({"error_description": "", "task_instance_id": None})
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task/<workflow_id>/set_resume_state", methods=["POST"])
def set_task_resume_state(workflow_id: int) -> Any:
    """An endpoint to set all tasks to a resumable state for a workflow.

    Conditioned on the workflow already being in an appropriate resume state.
    """
    data = cast(Dict, request.get_json())
    reset_if_running = bool(data["reset_if_running"])

    session = SessionLocal()
    with session.begin():
        # Ensure that the workflow is resumable
        # Necessary?
        workflow = session.execute(
            select(Workflow).where(Workflow.id == workflow_id)
        ).scalar()
        if not workflow.is_resumable:
            err_msg = (
                f"Workflow {workflow_id} is not resumable. Please "
                f"set the appropriate resume state."
            )
            resp = jsonify(err_msg=err_msg)
            resp.status_code = StatusCodes.OK
            return resp

        # Set task reset. If calling this bulk route, don't update any metadata besides what's
        # already bound in the database.

        # Logic: reset_if_running -> Reset all tasks not in "D" state
        # else, reset all tasks not in "D" or "R" state
        # for performance, also excclude TaskStatus.REGISTERING
        excluded_states = [TaskStatus.DONE, TaskStatus.REGISTERING]
        if not reset_if_running:
            excluded_states.append(TaskStatus.RUNNING)

        session.execute(
            update(Task)
            .where(Task.status.not_in(excluded_states), Task.workflow_id == workflow_id)
            .values(
                status=TaskStatus.REGISTERING, num_attempts=0, status_date=func.now()
            )
        )

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp



"""Routes for Clusters."""
from http import HTTPStatus as StatusCodes
from typing import Any

from flask import jsonify
from sqlalchemy import select

from jobmon.server.web.models.cluster import Cluster
from jobmon.server.web.models.cluster_type import ClusterType
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint


@blueprint.route("/cluster/<cluster_name>", methods=["GET"])
def get_cluster_by_name(cluster_name: str) -> Any:
    """Get the id, cluster_type_name and connection_parameters of a Cluster."""
    session = SessionLocal()
    with session.begin():
        select_stmt = (
            select(Cluster)
            .join(ClusterType, Cluster.cluster_type_id == ClusterType.id)
            .filter(Cluster.name == cluster_name)
        )
        cluster = session.execute(select_stmt).scalars().one_or_none()

    # send back json
    if cluster is None:
        resp = jsonify(cluster=None)
    else:
        resp = jsonify(cluster=cluster.to_wire_as_requested_by_client())
    resp.status_code = StatusCodes.OK
    return resp



"""Arg table in the database."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship

from jobmon.server.web.models import Base


class Arg(Base):
    """Arg table in the database."""

    __tablename__ = "arg"

    id = Column(Integer, primary_key=True)
    name = Column(String(255), unique=True)

    template_arg_map = relationship("TemplateArgMap", back_populates="argument")



"""Task Instance Status Database table."""
from sqlalchemy import Column, String
from sqlalchemy.orm import Session

from jobmon.core.constants import TaskStatus as Statuses
from jobmon.server.web.models import Base


class TaskStatus(Base):
    """The table in the database that holds on the possible statuses for a Task."""

    __tablename__ = "task_status"

    REGISTERING = Statuses.REGISTERING
    QUEUED = Statuses.QUEUED
    INSTANTIATING = Statuses.INSTANTIATING
    LAUNCHED = Statuses.LAUNCHED
    RUNNING = Statuses.RUNNING
    ERROR_RECOVERABLE = Statuses.ERROR_RECOVERABLE
    ADJUSTING_RESOURCES = Statuses.ADJUSTING_RESOURCES
    ERROR_FATAL = Statuses.ERROR_FATAL
    DONE = Statuses.DONE

    id = Column(String(1), primary_key=True)
    label = Column(String(150))
    description = Column(String(150))


def add_task_statuses(session: Session) -> None:
    """Populate the task_status table in the database."""
    statuses = [
        TaskStatus(
            id="A",
            label="ADJUSTING_RESOURCES",
            description="Task errored with a resource error, the resources will be "
            "adjusted before retrying.",
        ),
        TaskStatus(
            id="D",
            label="DONE",
            description="Task is Done, it ran successfully to completion; "
            "it has a TaskInstance that successfully completed.",
        ),
        TaskStatus(
            id="E",
            label="ERROR_RECOVERABLE",
            description="Task has errored out but has more attempts so it will be retried.",
        ),
        TaskStatus(
            id="F",
            label="ERROR_FATAL",
            description="Task errored out and has used all of the attempts, therefore has "
            "failed for this WorkflowRun. It can be resumed in a new WFR.",
        ),
        TaskStatus(
            id="G", label="REGISTERING", description="Task is bound to the database."
        ),
        TaskStatus(
            id="I", label="INSTANTIATING", description="Task is created within Jobmon."
        ),
        TaskStatus(
            id="O",
            label="LAUNCHED",
            description="Task instance submitted to the cluster normally, "
            "part of a Job Array.",
        ),
        TaskStatus(
            id="Q",
            label="QUEUED",
            description="Task's dependencies have successfully completed, task can be run "
            "when the scheduler is ready.",
        ),
        TaskStatus(
            id="R",
            label="RUNNING",
            description="Task is running on the specified distributor.",
        ),
    ]
    session.add_all(statuses)



"""Service to monitor and reap dead workflows."""
import logging
from time import sleep
from typing import Callable, List, Optional, Tuple

from jobmon.core.configuration import JobmonConfig
from jobmon.core.constants import WorkflowRunStatus
from jobmon.core.exceptions import ConfigError, InvalidResponse
from jobmon.core.requester import http_request_ok, Requester
from jobmon.server import __version__
from jobmon.server.workflow_reaper.notifiers import SlackNotifier
from jobmon.server.workflow_reaper.reaper_workflow_run import ReaperWorkflowRun

logger = logging.getLogger(__file__)


class WorkflowReaper(object):
    """Monitoring and reaping dead workflows."""

    _version = __version__

    # starting point of F-D inconsistency query
    _current_starting_row = 0

    _reaper_message = {
        WorkflowRunStatus.ERROR: (
            "{__version__} Workflow Reaper transitioned a Workflow to FAILED state and "
            "associated Workflow Run to ERROR state.\n"
            "Workflow ID: {workflow_id}\n"
            "Workflow Name: {workflow_name}\n"
            "Workflow Args: {workflow_args}\n"
            "WorkflowRun ID: {workflow_run_id}"
        ),
        WorkflowRunStatus.TERMINATED: (
            "{__version__} Workflow Reaper transitioned a Workflow to HALTED state and "
            "associated Workflow Run to TERMINATED state.\n"
            "Workflow ID: {workflow_id}\n"
            "Workflow Name: {workflow_name}\n"
            "Workflow Args: {workflow_args}\n"
            "WorkflowRun ID: {workflow_run_id}"
        ),
        WorkflowRunStatus.ABORTED: (
            "{__version__} Workflow Reaper transitioned a Workflow to ABORTED state and "
            "associated Workflow Run to ABORTED state.\n"
            "Workflow ID: {workflow_id}\n"
            "Workflow Name: {workflow_name}\n"
            "Workflow Args: {workflow_args}\n"
            "WorkflowRun ID: {workflow_run_id}"
        ),
    }

    def __init__(
        self,
        poll_interval_seconds: Optional[int] = None,
        requester: Optional[Requester] = None,
        wf_notification_sink: Optional[Callable[..., None]] = None,
    ) -> None:
        """Initializes WorkflowReaper class with specified poll interval and slack info.

        Args:
            poll_interval_seconds(int): how often the WorkflowReaper should check the
                database and reap workflows. Using seconds, rather than minutes, makes
                the tests run faster.
            requester (Requester): requester to communicate with Flask.
            wf_notification_sink (Callable): Slack notifier send().
        """
        config = JobmonConfig()

        # get poll interval from config
        if poll_interval_seconds is None:
            poll_interval_seconds = (
                config.get_int("reaper", "poll_interval_minutes") * 60
            )
        if requester is None:
            requester = Requester.from_defaults()
        if wf_notification_sink is None:
            try:
                wf_notifier = SlackNotifier()
                wf_notification_sink = wf_notifier.send
            except ConfigError:
                pass

        logger.info(
            f"WorkflowReaper initializing with: poll_interval_minutes={poll_interval_seconds},"
            f"requester_url={requester.url}"
        )

        self._poll_interval_seconds = poll_interval_seconds
        self._requester = requester
        self._wf_notification_sink = wf_notification_sink

    def monitor_forever(self) -> None:
        """The main part of the Worklow Reaper.

        Check if workflow runs should be in ABORTED, SUSPENDED, or ERROR state. Wait and do
        it again.
        """
        logger.info("Monitoring forever...")

        if self._wf_notification_sink is not None:
            self._wf_notification_sink(msg=f"Workflow Reaper v{__version__} is alive")
        try:
            while True:
                self._halted_state()
                self._aborted_state()
                self._error_state()
                # The chunk size for the _inconsistent_status query is small so that each
                # query takes 100-400 mS. Therefore run several, a few seconds apart. We want
                # to be able to clean the whole database every 12 hours, but also not lock
                # the database.
                for i in range(5):
                    self._inconsistent_status(100)
                    sleep(2)
                sleep(self._poll_interval_seconds)
        except RuntimeError as e:
            logger.debug(f"Error in monitor_forever() in workflow reaper: {e}")

    def _get_wf_name_args(self, workflow_id: int) -> Tuple[str, str]:
        """Return the workflow name and args associated with a specific workflow_id."""
        logger.info(
            f"Checking the DB for workflow name and args of WF_ID: {workflow_id}"
        )
        app_route = f"/workflow/{workflow_id}/workflow_name_and_args"
        return_code, result = self._requester.send_request(
            app_route=app_route,
            message={"workflow_id": workflow_id},
            request_type="get",
        )
        if http_request_ok(return_code) is False:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST "
                f"request through route {app_route}. Expected "
                f"code 200. Response content: {result}"
            )
        return result["workflow_name"], result["workflow_args"]

    def _get_lost_workflow_runs(self, status: List[str]) -> List[ReaperWorkflowRun]:
        """Return all workflows that are in a specific state."""
        logger.info(f"Checking the database for workflow runs of status: {status}")
        app_route = "/lost_workflow_run"
        return_code, result = self._requester.send_request(
            app_route=app_route,
            message={"status": status, "version": self._version},
            request_type="get",
        )
        if http_request_ok(return_code) is False:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST "
                f"request through route {app_route}. Expected "
                f"code 200. Response content: {result}"
            )
        workflow_runs = []
        for wfr in result["workflow_runs"]:
            workflow_runs.append(ReaperWorkflowRun.from_wire(wfr, self._requester))

        if workflow_runs:
            logger.info(f"Found workflow runs: {workflow_runs}")
        return workflow_runs

    def _halted_state(self) -> Optional[str]:
        """Check if a workflow_run needs to be transitioned to terminated state."""
        # Get workflow_runs in H and C state
        workflow_runs = self._get_lost_workflow_runs(["C", "H"])

        # Transition workflows to HALTED
        target_status = WorkflowRunStatus.TERMINATED
        messages = ""
        for wfr in workflow_runs:
            status = wfr.reap()
            if status == target_status and self._wf_notification_sink is not None:
                wf_name, wf_args = self._get_wf_name_args(wfr.workflow_id)
                message = self._reaper_message[status].format(
                    __version__=self._version,
                    workflow_id=wfr.workflow_id,
                    workflow_run_id=wfr.workflow_run_id,
                    workflow_name=wf_name,
                    workflow_args=wf_args,
                )
                self._wf_notification_sink(msg=message)
                messages += message
        return messages

    def _error_state(self) -> Optional[str]:
        """Get lost workflows and register them as error."""
        workflow_runs = self._get_lost_workflow_runs(["R"])

        # Transitions workflow to FAILED state and workflow run to ERROR
        target_status = WorkflowRunStatus.ERROR
        messages = ""
        for wfr in workflow_runs:
            status = wfr.reap()
            if status == target_status and self._wf_notification_sink is not None:
                wf_name, wf_args = self._get_wf_name_args(wfr.workflow_id)
                message = self._reaper_message[status].format(
                    __version__=self._version,
                    workflow_id=wfr.workflow_id,
                    workflow_run_id=wfr.workflow_run_id,
                    workflow_name=wf_name,
                    workflow_args=wf_args,
                )
                self._wf_notification_sink(msg=message)
                messages += message
        return messages

    def _aborted_state(self) -> Optional[str]:
        """Find workflows that should be in aborted state.

        Get all workflow runs in G state and validate if they should be in A state. Get all
        lost wfr in L state and set it to A
        """
        # Get all lost wfr in L
        workflow_runs = self._get_lost_workflow_runs(["L"])

        # Transitions workflow to A state and workflow run to A
        target_status = WorkflowRunStatus.ABORTED
        messages = ""
        for wfr in workflow_runs:
            status = wfr.reap()
            if status == target_status and self._wf_notification_sink is not None:
                wf_name, wf_args = self._get_wf_name_args(wfr.workflow_id)
                message = self._reaper_message[status].format(
                    __version__=self._version,
                    workflow_id=wfr.workflow_id,
                    workflow_run_id=wfr.workflow_run_id,
                    workflow_name=wf_name,
                    workflow_args=wf_args,
                )
                self._wf_notification_sink(msg=message)
                messages += message
        return messages

    def _inconsistent_status(self, step_size: int) -> None:
        """Find wf in F with all tasks in D and fix them."""
        logger.debug("Find wf in state F but all tasks in D and fix them.")

        app_route = (
            f"/workflow/{WorkflowReaper._current_starting_row}/fix_status_inconsistency"
        )
        return_code, result = self._requester.send_request(
            app_route=app_route,
            message={"increase_step": step_size},
            request_type="put",
        )
        if http_request_ok(return_code) is False:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST "
                f"request through route {app_route}. Expected "
                f"code 200. Response content: {result}"
            )
        WorkflowReaper._current_starting_row = int(result["wfid"])



"""Task Attribute Table."""
from typing import Any, Dict

from sqlalchemy import Column, ForeignKey, Integer, String

from jobmon.server.web.models import Base


class TaskAttribute(Base):
    """Task Attribute Table."""

    __tablename__ = "task_attribute"

    task_id = Column(Integer, ForeignKey("task.id"), primary_key=True)
    task_attribute_type_id = Column(
        Integer, ForeignKey("task_attribute_type.id"), primary_key=True
    )
    value = Column(String(255))

    @classmethod
    def from_wire(cls: Any, dct: Dict) -> Any:
        """Task Attribute object created from dict."""
        return cls(
            id=dct["task_attribute_id"],
            task_id=dct["task_id"],
            attribute_type_id=dct["attribute_type_id"],
            value=dct["value"],
        )

    def to_wire(self) -> Dict:
        """Task attribute attributes to dict."""
        return {
            "task_attribute_id": self.id,
            "task_id": self.task_id,
            "attribute_type_id": self.attribute_type_id,
            "value": self.value,
        }



"""Routes for TaskTemplate."""
from http import HTTPStatus as StatusCodes
import json
from typing import Any, Dict, List

from flask import jsonify, request
from flask_cors import cross_origin
import numpy as np
import polars as pl
import scipy.stats as st  # type:ignore
from sqlalchemy import select
from sqlalchemy.sql import func
import structlog

from jobmon.core.serializers import SerializeTaskTemplateResourceUsage
from jobmon.server.web.models.arg import Arg
from jobmon.server.web.models.array import Array
from jobmon.server.web.models.node import Node
from jobmon.server.web.models.node_arg import NodeArg
from jobmon.server.web.models.queue import Queue
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.task_instance import TaskInstance
from jobmon.server.web.models.task_instance_error_log import TaskInstanceErrorLog
from jobmon.server.web.models.task_resources import TaskResources
from jobmon.server.web.models.task_template import TaskTemplate
from jobmon.server.web.models.task_template_version import TaskTemplateVersion
from jobmon.server.web.models.workflow import Workflow
from jobmon.server.web.models.workflow_run import WorkflowRun
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.cli import blueprint
from jobmon.server.web.routes.cli.workflow import _cli_label_mapping
from jobmon.server.web.server_side_exception import InvalidUsage

# new structlog logger per flask request context. internally stored as flask.g.logger
logger = structlog.get_logger(__name__)


@blueprint.route("/get_task_template_version", methods=["GET"])
def get_task_template_version_for_tasks() -> Any:
    """Get the task_template_version_ids."""
    # parse args
    t_id = request.args.get("task_id")
    wf_id = request.args.get("workflow_id")
    # This route only accept one task id or one wf id;
    # If provided both, ignor wf id
    session = SessionLocal()
    with session.begin():
        if t_id:
            query_filter = [
                Task.id == t_id,
                Task.node_id == Node.id,
                Node.task_template_version_id == TaskTemplateVersion.id,
                TaskTemplateVersion.task_template_id == TaskTemplate.id,
            ]
            sql = select(
                TaskTemplateVersion.id,
                TaskTemplate.name,
            ).where(*query_filter)

        else:
            query_filter = [
                Task.workflow_id == wf_id,
                Task.node_id == Node.id,
                Node.task_template_version_id == TaskTemplateVersion.id,
                TaskTemplateVersion.task_template_id == TaskTemplate.id,
            ]
            sql = (
                select(
                    TaskTemplateVersion.id,
                    TaskTemplate.name,
                ).where(*query_filter)
            ).distinct()
        rows = session.execute(sql).all()
    column_names = ("id", "name")
    ttvis = [dict(zip(column_names, ti)) for ti in rows]

    resp = jsonify({"task_template_version_ids": ttvis})
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/get_requested_cores", methods=["GET"])
def get_requested_cores() -> Any:
    """Get the min, max, and arg of requested cores."""
    # parse args
    ttvis = request.args.get("task_template_version_ids")
    if ttvis is None:
        raise ValueError(
            "No task_template_version_ids returned in /get_requested_cores"
        )
    ttvis = [int(i) for i in ttvis[1:-1].split(",")]
    # null core should be treated as 1 instead of 0
    session = SessionLocal()
    with session.begin():
        query_filter = [
            TaskTemplateVersion.id.in_(ttvis),
            TaskTemplateVersion.id == Node.task_template_version_id,
            Task.node_id == Node.id,
            Task.task_resources_id == TaskResources.id,
        ]

        sql = select(TaskTemplateVersion.id, TaskResources.requested_resources).where(
            *query_filter
        )
    rows = session.execute(sql).all()
    column_names = ("id", "rr")
    rows = [dict(zip(column_names, ti)) for ti in rows]

    core_info = []
    if rows:
        result_dir: Dict = dict()
        for r in rows:
            # json loads hates single quotes
            j_str = r["rr"].replace("'", '"')
            j_dir = json.loads(j_str)
            core = 1 if "num_cores" not in j_dir.keys() else int(j_dir["num_cores"])
            if r["id"] in result_dir.keys():
                result_dir[r["id"]].append(core)
            else:
                result_dir[r["id"]] = [core]
        for k in result_dir.keys():
            item_min = int(np.min(result_dir[k]))
            item_max = int(np.max(result_dir[k]))
            item_mean = round(np.mean(result_dir[k]))
            core_info.append(
                {"id": k, "min": item_min, "max": item_max, "avg": item_mean}
            )

    resp = jsonify({"core_info": core_info})
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/get_most_popular_queue", methods=["GET"])
def get_most_popular_queue() -> Any:
    """Get the most popular queue of the task template."""
    # parse args
    ttvis = request.args.get("task_template_version_ids")
    if ttvis is None:
        raise ValueError(
            "No task_template_version_ids returned in /get_most_popular_queue."
        )
    ttvis = [int(i) for i in ttvis[1:-1].split(",")]
    session = SessionLocal()
    with session.begin():
        query_filter = [
            TaskTemplateVersion.id.in_(ttvis),
            TaskTemplateVersion.id == Node.task_template_version_id,
            Task.node_id == Node.id,
            TaskInstance.task_id == Task.id,
            TaskInstance.task_resources_id == TaskResources.id,
            TaskResources.queue_id.isnot(None),
        ]
        sql = select(TaskTemplateVersion.id, TaskResources.queue_id).where(
            *query_filter
        )

    rows = session.execute(sql).all()
    column_names = ("id", "queue_id")
    rows = [dict(zip(column_names, ti)) for ti in rows]
    # return a "standard" json format for cli routes
    queue_info = []
    if rows:
        result_dir: Dict = dict()
        for r in rows:
            ttvi = r["id"]
            q = r["queue_id"]
            if ttvi in result_dir.keys():
                if q in result_dir[ttvi].keys():
                    result_dir[ttvi][q] += 1
                else:
                    result_dir[ttvi][q] = 1
            else:
                result_dir[ttvi] = dict()
                result_dir[ttvi][q] = 1
        for ttvi in result_dir.keys():
            # assign to a variable to keep typecheck happy
            max_usage = 0
            for q in result_dir[ttvi].keys():
                if result_dir[ttvi][q] > max_usage:
                    popular_q = q
                    max_usage = result_dir[ttvi][q]
            # get queue name; and return queue id with it
            with session:
                query_filter = [Queue.id == popular_q]
                sql = select(Queue.name).where(*query_filter)
            popular_q_name = session.execute(sql).one()[0]
            queue_info.append(
                {"id": ttvi, "queue": popular_q_name, "queue_id": popular_q}
            )

    resp = jsonify({"queue_info": queue_info})
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task_template_resource_usage", methods=["POST"])
@cross_origin()
def get_task_template_resource_usage() -> Any:
    """Return the aggregate resource usage for a give TaskTemplate.

    Need to use cross_origin decorator when using the GUI to call a post route.
    This enables Cross Origin Resource Sharing (CORS) on the route. Default is
    most permissive settings.
    """
    data = request.get_json()
    try:
        task_template_version_id = data.pop("task_template_version_id")
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to /task_template_resource_usage", status_code=400
        ) from e

    workflows = data.pop("workflows", None)
    node_args = data.pop("node_args", None)
    ci = data.pop("ci", None)
    viz = bool(data.pop("viz", False))

    session = SessionLocal()
    with session.begin():
        query_filter = [
            TaskTemplateVersion.id == task_template_version_id,
            Task.status == "D",
            TaskInstance.status == "D",
            TaskTemplateVersion.id == Node.task_template_version_id,
            Node.id == Task.node_id,
            Task.id == TaskInstance.task_id,
        ]
        if workflows:
            query_filter += [
                TaskInstance.workflow_run_id == WorkflowRun.id,
                WorkflowRun.workflow_id == Workflow.id,
                Workflow.id.in_(workflows),
            ]
        sql = select(
            TaskInstance.wallclock, TaskInstance.maxrss, Node.id, Task.id
        ).where(*query_filter)
        rows = session.execute(sql).all()
        session.commit()
    column_names = ("r", "m", "node_id", "task_id")
    rows = [dict(zip(column_names, ti)) for ti in rows]
    result = []
    if rows:
        for r in rows:
            if r["r"] is None:
                r["r"] = 0
            if node_args:
                session = SessionLocal()
                with session.begin():
                    node_f = [NodeArg.arg_id == Arg.id, NodeArg.node_id == r["node_id"]]
                    node_s = select(Arg.name, NodeArg.val).where(*node_f)
                    node_rows = session.execute(node_s).all()
                    session.commit()
                _include = False
                for n in node_rows:
                    if not _include:
                        if n[0] in node_args.keys() and n[1] in node_args[n[0]]:
                            _include = True
                if _include:
                    result.append(r)
            else:
                result.append(r)

    if len(result) == 0:
        resource_usage = SerializeTaskTemplateResourceUsage.to_wire(
            None, None, None, None, None, None, None, None, None, None, None
        )
    else:
        runtimes = []
        mems = []
        for row in result:
            runtimes.append(int(row["r"]))
            mems.append(max(0, 0 if row["m"] is None else int(row["m"])))

        num_tasks = len(runtimes)
        # set 0 to NaN; thus, numpy ignores them
        if 0 in mems:
            mems.remove(0)
        if 0 in runtimes:
            runtimes.remove(0)
        if len(mems) > 0:
            min_mem = int(np.min(mems))
            max_mem = int(np.max(mems))
            mean_mem = round(float(np.mean(mems)), 2)
            median_mem = round(float(np.percentile(mems, 50)), 2)
        else:
            min_mem = 0
            max_mem = 0
            mean_mem = 0
            median_mem = 0
        if len(runtimes) > 0:
            min_runtime = int(np.min(runtimes))
            max_runtime = int(np.max(runtimes))
            mean_runtime = round(float(np.mean(runtimes)), 2)
            median_runtime = round(float(np.percentile(runtimes, 50)), 2)
        else:
            min_runtime = 0
            max_runtime = 0
            mean_runtime = 0
            median_runtime = 0

        if ci is None:
            ci_mem = [None, None]
            ci_runtime = [None, None]
        else:
            try:
                ci = float(ci)

                def _calculate_ci(d: List, ci: float) -> List[Any]:
                    interval = st.t.interval(
                        alpha=ci, df=len(d) - 1, loc=np.mean(d), scale=st.sem(d)
                    )
                    return [round(float(interval[0]), 2), round(float(interval[1]), 2)]

                if len(mems) > 0:
                    ci_mem = _calculate_ci(mems, ci)
                else:
                    ci_mem = [None, None]
                if len(runtimes) > 0:
                    ci_runtime = _calculate_ci(runtimes, ci)
                else:
                    ci_runtime = [None, None]

            except ValueError as e:
                logger.warn(
                    f"Unable to convert {ci} to float. Use None. Exception: {str(e)}"
                )
                ci_mem = [None, None]
                ci_runtime = [None, None]

        resource_usage = SerializeTaskTemplateResourceUsage.to_wire(
            num_tasks,
            min_mem,
            max_mem,
            mean_mem,
            min_runtime,
            max_runtime,
            mean_runtime,
            median_mem,
            median_runtime,
            ci_mem,
            ci_runtime,
        )

    if viz:
        resource_usage += (result,)

    resp = jsonify(resource_usage)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow_tt_status_viz/<workflow_id>", methods=["GET"])
def get_workflow_tt_status_viz(workflow_id: int) -> Any:
    """Get the status of the workflows for GUI."""
    # return DS
    return_dic: Dict[int, Any] = dict()

    session = SessionLocal()
    with session.begin():
        # user subquery as the Array table has to be joined on two columns
        sub_query = (
            select(
                Array.id, Array.task_template_version_id, Array.max_concurrently_running
            ).where(Array.workflow_id == workflow_id)
        ).subquery()
        join_table = (
            Task.__table__.join(Node, Task.node_id == Node.id)
            .join(
                TaskTemplateVersion,
                Node.task_template_version_id == TaskTemplateVersion.id,
            )
            .join(
                TaskTemplate,
                TaskTemplateVersion.task_template_id == TaskTemplate.id,
            )
            # Arrays were introduced in 3.1.0, hence the outer-join for 3.0.* workflows
            .join(
                sub_query,
                sub_query.c.task_template_version_id == TaskTemplateVersion.id,
                isouter=True,
            )
        )
        # Order by the task submitted date in each task template
        sql = (
            select(
                TaskTemplate.id,
                TaskTemplate.name,
                Task.id,
                Task.status,
                sub_query.c.max_concurrently_running,
                TaskTemplateVersion.id,
            )
            .select_from(join_table)
            .where(Task.workflow_id == workflow_id)
            .order_by(Task.id)
        )
        # For performance reasons, use STRAIGHT_JOIN to set the join order. If not set,
        # the optimizer may choose a suboptimal execution plan for large datasets.
        # Has to be conditional since not all database engines support STRAIGHT_JOIN.
        if SessionLocal.bind.dialect.name == "mysql":
            sql = sql.prefix_with("STRAIGHT_JOIN")
        rows = session.execute(sql).all()
        session.commit()

    # Get min, max, avg for each task template in `workflow_id`
    with session.begin():
        join_table = (
            Task.__table__.join(Node, Task.node_id == Node.id)
            .join(
                TaskTemplateVersion,
                Node.task_template_version_id == TaskTemplateVersion.id,
            )
            .join(
                TaskTemplate,
                TaskTemplateVersion.task_template_id == TaskTemplate.id,
            )
        )
        sql = (
            select(
                TaskTemplate.id.label("task_template_id"),
                TaskTemplate.name.label("task_template_name"),
                func.min(Task.num_attempts).label("min"),
                func.max(Task.num_attempts).label("max"),
                func.avg(Task.num_attempts).label("mean"),
            )
            .select_from(join_table)
            .where(Task.workflow_id == workflow_id)
            .group_by(TaskTemplate.id)
        )
        if SessionLocal.bind.dialect.name == "mysql":
            sql = sql.prefix_with("STRAIGHT_JOIN")
        attempts = session.execute(sql).all()

    attempts = {attempt[0]: attempt for attempt in attempts}

    for r in rows:
        # Avoiding magic numbers
        task_template_id: str = r[0]
        task_template_name: str = r[1]
        task_status: str = r[3]
        max_concurrently = r[4]
        task_template_version_id: int = int(r[5])

        if int(task_template_id) in return_dic.keys():
            pass
        else:
            attempt = attempts.get(task_template_id)
            *_, min_, max_, mean = attempt

            return_dic[int(task_template_id)] = {
                "id": int(task_template_id),
                "name": task_template_name,
                "tasks": 0,
                "PENDING": 0,
                "SCHEDULED": 0,
                "RUNNING": 0,
                "DONE": 0,
                "FATAL": 0,
                "MAXC": 0,
                "num_attempts_min": min_,
                "num_attempts_max": max_,
                "num_attempts_avg": mean,
                "task_template_version_id": task_template_version_id,
            }
        return_dic[int(task_template_id)]["tasks"] += 1
        return_dic[int(task_template_id)][_cli_label_mapping[task_status]] += 1
        return_dic[int(task_template_id)]["MAXC"] = (
            max_concurrently if max_concurrently is not None else "NA"
        )
    resp = jsonify(return_dic)
    resp.status_code = 200
    return resp


@blueprint.route("/tt_error_log_viz/<wf_id>/<tt_id>", methods=["GET"])
def get_tt_error_log_viz(tt_id: int, wf_id: int) -> Any:
    """Get the error logs for a task template id for GUI."""
    # return DS
    return_list: List[Any] = []

    session = SessionLocal()
    with session.begin():
        query_filter = [
            TaskTemplateVersion.task_template_id == tt_id,
            Node.task_template_version_id == TaskTemplateVersion.id,
            Task.node_id == Node.id,
            Task.workflow_id == wf_id,
            TaskInstance.task_id == Task.id,
            TaskInstanceErrorLog.task_instance_id == TaskInstance.id,
        ]

        sql = (
            select(
                Task.id,
                TaskInstance.id,
                TaskInstanceErrorLog.id,
                TaskInstanceErrorLog.error_time,
                TaskInstanceErrorLog.description,
                TaskInstance.stderr_log,
            )
            .where(*query_filter)
            .order_by(TaskInstanceErrorLog.id.desc())
        )
        # For performance reasons, use STRAIGHT_JOIN to set the join order. If not set,
        # the optimizer may choose a suboptimal execution plan for large datasets.
        # Has to be conditional since not all database engines support STRAIGHT_JOIN.
        if SessionLocal.bind.dialect.name == "mysql":
            sql = sql.prefix_with("STRAIGHT_JOIN")
        rows = session.execute(sql).all()
        session.commit()
    for r in rows:
        # dict: {<error log id>: [<tid>, <tiid>, <error time>, <error log>}
        return_list.append(
            {
                "task_id": r[0],
                "task_instance_id": r[1],
                "task_instance_err_id": r[2],
                "error_time": r[3],
                "error": r[4],
                "task_instance_stderr_log": r[5],
            }
        )

    # Create Polars DataFrame with the errors, initializing most_recent_attempt to False
    error_schema = {
        "error": pl.Utf8,
        "error_time": pl.Datetime,
        "task_id": pl.Int32,
        "task_instance_err_id": pl.Int32,
        "task_instance_id": pl.Int32,
        "task_instance_stderr_log": pl.Utf8,
    }

    errors_df = pl.DataFrame(return_list, schema=error_schema)
    errors_df = (
        errors_df.lazy()
        .with_columns(pl.lit(False).alias("most_recent_attempt"))
        .collect()
    )

    # Create Polars DataFrame of the most recent attempts
    errors_most_recent_df = (
        errors_df.lazy()
        .groupby("task_id")
        .agg([pl.col("task_instance_id").max()])
        .with_columns(pl.lit(True).alias("most_recent_attempt"))
        .collect()
    )

    #  Update original DF with the second (join + coalesce)
    errors_df = (
        errors_df.lazy()
        .update(errors_most_recent_df.lazy(), on=["task_instance_id"], how="left")
        .collect()
    )

    resp = jsonify(errors_df.to_dicts())

    resp.status_code = 200
    return resp



from http import HTTPStatus
from typing import Any

from flask import jsonify, request
from sqlalchemy import select

from jobmon.server.web.models.array import Array
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.task_instance import TaskInstance
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.cli import blueprint


@blueprint.route("/array/<workflow_id>/get_array_tasks")
def get_array_task_instances(workflow_id: int) -> Any:
    """Return error/output filepaths for task instances filtered by array name.

    The user can also optionally filter by job name as well.

    To avoid overly-large returned results, the user must also pass in a workflow ID.
    """
    data = request.args
    array_name = data.get("array_name")
    job_name = data.get("job_name")
    limit = data.get("limit", 5)

    query_filters = [
        Task.workflow_id == workflow_id,
        TaskInstance.task_id == Task.id,
        Task.array_id == Array.id,
    ]

    if array_name:
        query_filters.append(Array.name == array_name)

    if job_name:
        query_filters.append(Task.name == job_name)

    session = SessionLocal()
    with SessionLocal.begin():
        select_stmt = (
            select(
                Task.id,
                Task.name,
                Array.name,
                TaskInstance.id,
                TaskInstance.stdout,
                TaskInstance.stderr,
            )
            .where(*query_filters)
            .limit(limit)
        )
        result = session.execute(select_stmt).all()

    column_names = (
        "TASK_ID",
        "TASK_NAME",
        "ARRAY_NAME",
        "TASK_INSTANCE_ID",
        "OUTPUT_PATH",
        "ERROR_PATH",
    )
    resp = jsonify(array_tasks=[dict(zip(column_names, ti)) for ti in result])
    resp.status_code = HTTPStatus.OK
    return resp



"""Routes for DAGs."""
from http import HTTPStatus as StatusCodes
from typing import Any, cast, Dict

from flask import jsonify, request
import sqlalchemy
from sqlalchemy import insert, select, update
from sqlalchemy.sql import func
import structlog

from jobmon.server.web.models.dag import Dag
from jobmon.server.web.models.edge import Edge
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint
from jobmon.server.web.server_side_exception import InvalidUsage


# new structlog logger per flask request context. internally stored as flask.g.logger
logger = structlog.get_logger(__name__)


@blueprint.route("/dag", methods=["POST"])
def add_dag() -> Any:
    """Add a new dag to the database.

    Args:
        dag_hash: unique identifier of the dag, included in route
    """
    data = cast(Dict, request.get_json())

    # add dag
    dag_hash = data.pop("dag_hash")
    structlog.contextvars.bind_contextvars(dag_hash=str(dag_hash))
    logger.info(f"Add dag:{dag_hash}")

    session = SessionLocal()

    try:
        with session.begin():
            dag = Dag(hash=dag_hash)
            session.add(dag)

    except sqlalchemy.exc.IntegrityError:
        with session.begin():
            select_stmt = select(Dag).filter(Dag.hash == dag_hash)
            dag = session.execute(select_stmt).scalar_one()

    # return result
    resp = jsonify(dag_id=dag.id, created_date=dag.created_date)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/dag/<dag_id>/edges", methods=["POST"])
def add_edges(dag_id: int) -> Any:
    """Add edges to the edge table."""
    structlog.contextvars.bind_contextvars(dag_id=dag_id)
    logger.info(f"Add edges for dag {dag_id}")
    try:
        data = cast(Dict, request.get_json())
        edges_to_add = data.pop("edges_to_add")
        mark_created = bool(data.pop("mark_created"))
    except KeyError as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    # add dag and cast types
    for edges in edges_to_add:
        edges["dag_id"] = dag_id
        if len(edges["upstream_node_ids"]) == 0:
            edges["upstream_node_ids"] = None
        else:
            edges["upstream_node_ids"] = str(edges["upstream_node_ids"])

        if len(edges["downstream_node_ids"]) == 0:
            edges["downstream_node_ids"] = None
        else:
            edges["downstream_node_ids"] = str(edges["downstream_node_ids"])

    # Bulk insert the nodes and node args with raw SQL, for performance. Ignore duplicate
    # keys
    session = SessionLocal()
    with session.begin():
        insert_stmt = insert(Edge).values(edges_to_add)
        if SessionLocal.bind.dialect.name == "mysql":
            insert_stmt = insert_stmt.prefix_with("IGNORE")
        if SessionLocal.bind.dialect.name == "sqlite":
            insert_stmt = insert_stmt.prefix_with("OR IGNORE")
        session.execute(insert_stmt)
        session.flush()

        if mark_created:
            update_stmt = (
                update(Dag).where(Dag.id == dag_id).values(created_date=func.now())
            )
            session.execute(update_stmt)

    # return result
    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp



"""Routes used by task instances on worker nodes."""
from http import HTTPStatus as StatusCodes
import os
from typing import Any

from flask import current_app, jsonify
from sqlalchemy import func, select
from sqlalchemy import orm
from structlog import get_logger

from jobmon.server.web import session_factory

# scoped session associated with the current thread
SessionLocal = orm.scoped_session(session_factory)


logger = get_logger(__name__)


# ############################ SHARED LANDING ROUTES ##########################################
def is_alive() -> Any:
    """Action that sends a response to the requester indicating that responder is listening."""
    logger.info(f"{os.getpid()}: {current_app.__class__.__name__} received is_alive?")
    resp = jsonify(msg="Yes, I am alive")
    resp.status_code = StatusCodes.OK
    return resp


def _get_time() -> str:
    with SessionLocal() as session:
        db_time = session.execute(select(func.now())).scalar()
        str_time = db_time.strftime("%Y-%m-%d %H:%M:%S")
    return str_time


def get_pst_now() -> Any:
    """Get the time from the database."""
    time = _get_time()
    resp = jsonify(time=time)
    resp.status_code = StatusCodes.OK
    return resp


def health() -> Any:
    """Test connectivity to the database.

    Return 200 if everything is OK. Defined in each module with a different route, so it can
    be checked individually.
    """
    _get_time()
    resp = jsonify(status="OK")
    resp.status_code = StatusCodes.OK
    return resp


# ############################ TESTING ROUTES ################################################
def test_route() -> None:
    """Test route to force a 500 error."""
    session = SessionLocal()
    with session.begin():
        session.execute("SELECT * FROM blip_bloop_table").all()
        session.commit()



"""Queue Table in the Database."""
from sqlalchemy import Column, ForeignKey, Integer, select, String, UniqueConstraint
from sqlalchemy.orm import relationship, Session

from jobmon.core.serializers import SerializeQueue
from jobmon.server.web.models import Base
from jobmon.server.web.models.cluster import Cluster


class Queue(Base):
    """Queue Table in the Database."""

    __tablename__ = "queue"

    def to_wire_as_requested_by_client(self) -> tuple:
        """Serialize cluster object."""
        return SerializeQueue.to_wire(self.id, self.name, self.parameters)

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    cluster_id = Column(Integer, ForeignKey("cluster.id"))
    parameters = Column(String(2500))

    # ORM relationships
    cluster = relationship("Cluster", back_populates="queues")

    __table_args__ = (
        UniqueConstraint("name", "cluster_id", name="uc_name_cluster_id"),
    )


def add_queues(session: Session) -> None:
    """Populate the queue table in the database."""
    for cluster_name in ["dummy", "sequential", "multiprocess"]:
        cluster = (
            session.execute(select(Cluster).where(Cluster.name == cluster_name))
            .scalars()
            .one()
        )
        if cluster_name == "multiprocess":
            parameters = '{"cores": (1,20)}'
        else:
            parameters = "{}"
        session.add(Queue(name="null.q", cluster_id=cluster.id, parameters=parameters))



"""Node arg db table."""
from sqlalchemy import Column, Integer, String

from jobmon.server.web.models import Base


class NodeArg(Base):
    """Node arg db table."""

    __tablename__ = "node_arg"

    node_id = Column(Integer, primary_key=True)
    arg_id = Column(Integer, primary_key=True)
    val = Column(String(2048))



"""Workflow Database Table."""
import datetime
from typing import Tuple

from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Text, VARCHAR
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func
import structlog

from jobmon.core.exceptions import InvalidStateTransition
from jobmon.core.serializers import SerializeDistributorWorkflow
from jobmon.server.web.models import Base
from jobmon.server.web.models.workflow_run import WorkflowRun
from jobmon.server.web.models.workflow_run_status import WorkflowRunStatus
from jobmon.server.web.models.workflow_status import WorkflowStatus


logger = structlog.get_logger(__name__)


class Workflow(Base):
    """Workflow Database Table."""

    __tablename__ = "workflow"

    def to_wire_as_distributor_workflow(self) -> tuple:
        """Serialize workflow object."""
        serialized = SerializeDistributorWorkflow.to_wire(
            workflow_id=self.id,
            dag_id=self.dag_id,
            max_concurrently_running=self.max_concurrently_running,
        )
        return serialized

    id = Column(Integer, primary_key=True)
    tool_version_id = Column(Integer, ForeignKey("tool_version.id"))
    dag_id = Column(Integer, ForeignKey("dag.id"))
    workflow_args_hash = Column(VARCHAR(50), index=True)
    task_hash = Column(VARCHAR(50), index=True)
    description = Column(Text)
    name = Column(String(150))
    workflow_args = Column(Text)
    max_concurrently_running = Column(Integer)
    status = Column(
        String(1),
        ForeignKey("workflow_status.id"),
        default=WorkflowStatus.REGISTERING,
    )
    created_date = Column(DateTime, default=None)
    status_date = Column(DateTime, default=func.now())

    dag = relationship("Dag", back_populates="workflow", lazy=True)
    workflow_runs = relationship("WorkflowRun", back_populates="workflow", lazy=True)

    valid_transitions = [
        # normal progression from registered to a workflow run has been fully bound
        (WorkflowStatus.REGISTERING, WorkflowStatus.QUEUED),
        # workflow encountered an error before a workflow run was created.
        (WorkflowStatus.REGISTERING, WorkflowStatus.ABORTED),
        # a workflow aborted during task creation. new workflow launched, found
        # existing workflow id and is creating a new workflow run
        (WorkflowStatus.ABORTED, WorkflowStatus.REGISTERING),
        # new workflow run created that resumes old failed workflow run
        (WorkflowStatus.FAILED, WorkflowStatus.REGISTERING),
        # new workflow run created that resumes old suspended workflow run
        (WorkflowStatus.HALTED, WorkflowStatus.REGISTERING),
        # Workflow is instantiating
        (WorkflowStatus.QUEUED, WorkflowStatus.INSTANTIATING),
        # Workflow fails before it's instantiated
        (WorkflowStatus.QUEUED, WorkflowStatus.FAILED),
        # workflow is launched. normal happy path
        (WorkflowStatus.INSTANTIATING, WorkflowStatus.LAUNCHED),
        # workflow can't launch for some reason. TODO: Implement triaging
        (WorkflowStatus.INSTANTIATING, WorkflowStatus.FAILED),
        (WorkflowStatus.LAUNCHED, WorkflowStatus.FAILED),
        # workflow runs. normal happy path
        (WorkflowStatus.LAUNCHED, WorkflowStatus.RUNNING),
        # workflow run was running and then got moved to a resume state
        (WorkflowStatus.RUNNING, WorkflowStatus.HALTED),
        # workflow run was bound, and then got moved to a resume state
        (WorkflowStatus.QUEUED, WorkflowStatus.HALTED),
        # workflow run was running and then completed successfully
        (WorkflowStatus.RUNNING, WorkflowStatus.DONE),
        # workflow run was running and then failed with an error
        (WorkflowStatus.RUNNING, WorkflowStatus.FAILED),
    ]

    untimely_transitions = [(WorkflowStatus.REGISTERING, WorkflowStatus.REGISTERING)]

    def transition(self, new_state: str) -> None:
        """Transition the state of the workflow."""
        # bind_to_logger(workflow_id=self.id)
        logger.info(f"Transitioning workflow_id from {self.status} to {new_state}")
        if self._is_timely_transition(new_state):
            self._validate_transition(new_state)
            self.status = new_state
            self.status_date = func.now()
        logger.info(f"WorkflowStatus is now {self.status}")

    def _validate_transition(self, new_state: str) -> None:
        """Ensure the Job state transition is valid."""
        if (self.status, new_state) not in self.valid_transitions:
            raise InvalidStateTransition("Workflow", self.id, self.status, new_state)

    def _is_timely_transition(self, new_state: str) -> bool:
        """Check if the transition is invalid due to a race condition."""
        if (self.status, new_state) in self.untimely_transitions:
            return False
        else:
            return True

    def link_workflow_run(
        self, workflow_run: WorkflowRun, next_report_increment: float
    ) -> Tuple:
        """Link a workflow run to this workflow."""
        # bind_to_logger(workflow_id=self.id)
        logger.info(f"Linking WorkflowRun {workflow_run.id} to Workflow")
        linked_wfr = [
            wfr.status == WorkflowRunStatus.LINKING for wfr in self.workflow_runs
        ]

        if not any(linked_wfr) and self.ready_to_link:
            workflow_run.heartbeat(next_report_increment, WorkflowRunStatus.LINKING)
            current_wfr = [(workflow_run.id, workflow_run.status)]
        # active workflow run, don't bind.
        elif not any(linked_wfr) and not self.ready_to_link:
            # return currently alive workflow instead
            current_wfr = [
                (wfr.id, wfr.status) for wfr in self.workflow_runs if wfr.is_alive
            ]
        # currently linked workflow run
        else:
            current_wfr = [(wfr.id, wfr.status) for wfr in self.workflow_runs]

        if current_wfr:
            return current_wfr[0]
        else:
            return ()

    def reset(self, current_time: datetime.datetime) -> None:
        """Set a workflow to a resumable state."""
        # Terminate the first existing workflowrun that is in cold resume state
        for workflow_run in self.workflow_runs:
            if workflow_run.terminable(current_time=current_time):
                workflow_run.transition(WorkflowRunStatus.TERMINATED)
                break

        # Bypass the FSM, since we should be able to reset workflows in weird states.
        self.status = WorkflowStatus.REGISTERING
        self.status_date = func.now()

    def resume(self, reset_running_jobs: bool) -> None:
        """Resume a workflow."""
        # bind_to_logger(workflow_id=self.id)
        logger.info("Resume workflow")
        for workflow_run in self.workflow_runs:
            if workflow_run.is_active:
                if reset_running_jobs:
                    workflow_run.cold_reset()
                else:
                    workflow_run.hot_reset()
                break

    @property
    def ready_to_link(self) -> bool:
        """Is this workflow able to link a new workflow run."""
        return (
            self.status
            not in [
                WorkflowStatus.QUEUED,
                WorkflowStatus.RUNNING,
                WorkflowStatus.DONE,
            ]
            and self.is_resumable
        )

    @property
    def is_resumable(self) -> bool:
        """Is this workflow resumable."""
        wfrs_active = any([wfr.is_alive for wfr in self.workflow_runs])
        done_binding = self.created_date is not None
        return done_binding and not wfrs_active



"""DAG Database Table."""
from sqlalchemy import Column, DateTime, Integer, VARCHAR
from sqlalchemy.orm import relationship

from jobmon.server.web.models import Base


class Dag(Base):
    """DAG Database Table."""

    __tablename__ = "dag"

    id = Column(Integer, primary_key=True)
    hash = Column(VARCHAR(50), unique=True)
    created_date = Column(DateTime, default=None)

    workflow = relationship("Workflow", back_populates="dag", lazy=True)



"""Routes for WorkflowRuns."""
from http import HTTPStatus as StatusCodes
from typing import Any, cast, Dict, List

from flask import jsonify, request
from sqlalchemy import case, func, insert, select, update
import structlog

from jobmon.core import constants
from jobmon.core.exceptions import InvalidStateTransition
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.task_instance import TaskInstance
from jobmon.server.web.models.task_instance_error_log import TaskInstanceErrorLog
from jobmon.server.web.models.workflow import Workflow
from jobmon.server.web.models.workflow_run import WorkflowRun
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint
from jobmon.server.web.server_side_exception import InvalidUsage


logger = structlog.get_logger(__name__)


@blueprint.route("/workflow_run", methods=["POST"])
def add_workflow_run() -> Any:
    """Add a workflow run to the db."""
    try:
        data = cast(Dict, request.get_json())
        workflow_id = int(data["workflow_id"])
        user = data["user"]
        jobmon_version = data["jobmon_version"]
        next_heartbeat = float(data["next_report_increment"])

        structlog.contextvars.bind_contextvars(workflow_id=workflow_id)
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    logger.info(f"Add wfr for workflow_id:{workflow_id}.")

    session = SessionLocal()

    with session.begin():
        workflow = session.execute(
            select(Workflow).where(Workflow.id == workflow_id)
        ).scalar()
        err_msg = ""
        if not workflow:
            # Binding to a non-existent workflow, exit early
            err_msg = f"No workflow exists for ID {workflow_id}"
            resp = jsonify(workflow_run_id=None, err_msg=err_msg)
            resp.status_code = StatusCodes.OK
            return resp

        workflow_run = WorkflowRun(
            workflow_id=workflow_id,
            user=user,
            jobmon_version=jobmon_version,
            status=constants.WorkflowRunStatus.REGISTERED,
        )
        session.add(workflow_run)
        session.flush()

        session.refresh(workflow, with_for_update=True)
        # Transition to linking state, with_for_update claims a lock on the workflow
        # Any other actively linking workflows will return the incorrect workflow run id

        active_workflow_run = workflow.link_workflow_run(workflow_run, next_heartbeat)
        session.flush()

        try:
            if active_workflow_run[0] != workflow_run.id:
                err_msg = (
                    f"WorkflowRun {active_workflow_run[0]} is currently"
                    f"linking, WorkflowRun {workflow_run.id} will be aborted."
                )
        except IndexError:
            # Raised if the workflow is not resume-able, without any active workflowruns
            # Unlikely to be raised
            err_msg = f"Workflow {workflow_id} is not in a resume-able state"

    session.commit()
    if err_msg:
        logger.warning(f"Possible race condition in adding workflowrun: {err_msg}")
        resp = jsonify(workflow_run_id=None, err_msg=err_msg)
    else:
        logger.info(f"Add workflow_run:{workflow_run.id} for workflow.")
        resp = jsonify(workflow_run_id=workflow_run.id, status=workflow_run.status)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/workflow_run/<workflow_run_id>/terminate_task_instances", methods=["PUT"]
)
def terminate_workflow_run(workflow_run_id: int) -> Any:
    """Terminate a workflow run and get its tasks in order."""
    structlog.contextvars.bind_contextvars(workflow_run_id=workflow_run_id)
    logger.info("Terminate workflow_run")
    try:
        workflow_run_id = int(workflow_run_id)
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    session = SessionLocal()
    with session.begin():
        select_stmt = select(WorkflowRun).where(WorkflowRun.id == workflow_run_id)
        workflow_run = session.execute(select_stmt).scalars().one()

        if workflow_run.status == constants.WorkflowRunStatus.HOT_RESUME:
            task_states = [constants.TaskStatus.LAUNCHED]
        else:
            task_states = [constants.TaskStatus.LAUNCHED, constants.TaskStatus.RUNNING]

        insert_error_log_stmt = insert(TaskInstanceErrorLog).from_select(
            ["task_instance_id", "description", "error_time"],
            select(
                TaskInstance.id,
                (
                    "Workflow resume requested. Setting to K from status of: "
                    + TaskInstance.status
                ),
                func.now(),
            ).where(
                TaskInstance.workflow_run_id == workflow_run_id,
                TaskInstance.status == constants.TaskInstanceStatus.KILL_SELF,
            ),
        )
        session.execute(insert_error_log_stmt)

        workflow_id = workflow_run.workflow_id
        update_task_instance_stmt = (
            update(TaskInstance)
            .where(
                TaskInstance.id.in_(
                    select(TaskInstance.id).where(
                        TaskInstance.workflow_run_id == WorkflowRun.id,
                        TaskInstance.task_id.in_(
                            select(Task.id).where(
                                Task.workflow_id == workflow_id,
                                Task.status.in_(task_states),
                            )
                        ),
                    )
                )
            )
            .values(
                status=constants.TaskInstanceStatus.KILL_SELF, status_date=func.now()
            )
            .execution_options(synchronize_session=False)
        )

        session.execute(update_task_instance_stmt)

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow_run/<workflow_run_id>/log_heartbeat", methods=["POST"])
def log_workflow_run_heartbeat(workflow_run_id: int) -> Any:
    """Log a heartbeat for the workflow run to show that the client side is still alive."""
    structlog.contextvars.bind_contextvars(workflow_run_id=workflow_run_id)
    try:
        workflow_run_id = int(workflow_run_id)
        data = cast(Dict, request.get_json())
        next_report_increment = data["next_report_increment"]
        status = data["status"]
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    logger.debug(f"WFR {workflow_run_id} heartbeat data")

    session = SessionLocal()
    with session.begin():
        select_stmt = select(WorkflowRun).where(WorkflowRun.id == workflow_run_id)
        workflow_run = session.execute(select_stmt).scalars().one()

        try:
            workflow_run.heartbeat(next_report_increment, status)
            logger.debug(f"wfr {workflow_run_id} heartbeat confirmed")
        except InvalidStateTransition as e:
            logger.debug(f"wfr {workflow_run_id} heartbeat rolled back, reason: {e}")

    resp = jsonify(status=str(workflow_run.status))
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow_run/<workflow_run_id>/update_status", methods=["PUT"])
def log_workflow_run_status_update(workflow_run_id: int) -> Any:
    """Update the status of the workflow run."""
    structlog.contextvars.bind_contextvars(workflow_run_id=workflow_run_id)
    try:
        workflow_run_id = int(workflow_run_id)
        data = cast(Dict, request.get_json())
        status = data["status"]
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    logger.info(f"Log status update for workflow_run_id:{workflow_run_id}.")

    session = SessionLocal()
    with session.begin():
        select_stmt = select(WorkflowRun).where(WorkflowRun.id == workflow_run_id)
        workflow_run = session.execute(select_stmt).scalars().one()

        try:
            workflow_run.transition(status)
        except InvalidStateTransition as e:
            logger.warning(e)

        # Return the status
        status = workflow_run.status

    resp = jsonify(status=status)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/workflow_run/<workflow_run_id>/sync_status", methods=["POST"])
def task_instances_status_check(workflow_run_id: int) -> Any:
    """Sync status of given task intance IDs."""
    structlog.contextvars.bind_contextvars(workflow_run_id=workflow_run_id)
    try:
        workflow_run_id = int(workflow_run_id)
        data = cast(Dict, request.get_json())
        task_instance_ids = data["task_instance_ids"]
        status = data["status"]
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    session = SessionLocal()
    with session.begin():
        # get time from db
        db_time = session.execute(select(func.now())).scalar()
        str_time = db_time.strftime("%Y-%m-%d %H:%M:%S")

        where_clause = [TaskInstance.workflow_run_id == workflow_run_id]
        if len(task_instance_ids) > 0:
            # Filters for
            # 1) instances that have changed out of the declared status
            # 2) instances that have changed into the declared status
            where_clause.append(
                (
                    TaskInstance.id.in_(task_instance_ids)
                    & (TaskInstance.status != status)
                )
                | (
                    TaskInstance.id.notin_(task_instance_ids)
                    & (TaskInstance.status == status)
                )
            )
        else:
            where_clause.append(TaskInstance.status == status)

        select_stmt = (
            select(TaskInstance.status, func.group_concat(TaskInstance.id))
            .where(*where_clause)
            .group_by(TaskInstance.status)
        )

        return_dict: Dict[str, List[int]] = {}
        for row in session.execute(select_stmt):
            return_dict[row[0]] = [int(tid) for tid in row[1].split(",")]

    resp = jsonify(status_updates=return_dict, time=str_time)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/workflow_run/<workflow_run_id>/set_status_for_triaging", methods=["POST"]
)
def set_status_for_triaging(workflow_run_id: int) -> Any:
    """Two triaging related status sets.

    Query all task instances that are submitted to distributor or running which haven't
    reported as alive in the allocated time, and set them for Triaging(from Running)
    and Kill_self(from Launched).
    """
    structlog.contextvars.bind_contextvars(workflow_run_id=workflow_run_id)
    try:
        workflow_run_id = int(workflow_run_id)
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e
    logger.info(f"Set to triaging those overdue tis for wfr {workflow_run_id}")

    session = SessionLocal()
    with session.begin():
        update_stmt = (
            update(TaskInstance)
            .where(
                TaskInstance.workflow_run_id == workflow_run_id,
                TaskInstance.status.in_(
                    [
                        constants.TaskInstanceStatus.LAUNCHED,
                        constants.TaskInstanceStatus.RUNNING,
                    ]
                ),
                TaskInstance.report_by_date <= func.now(),
            )
            .values(
                status=case(
                    (
                        TaskInstance.status == constants.TaskInstanceStatus.RUNNING,
                        constants.TaskInstanceStatus.TRIAGING,
                    ),
                    else_=constants.TaskInstanceStatus.KILL_SELF,
                )
            )
            .execution_options(synchronize_session=False)
        )
        session.execute(update_stmt)
    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp



"""Places to notify upon certain events (ex. slack to notify of unhealthy workflow)."""
import logging
from typing import Optional

import requests

from jobmon.core.configuration import JobmonConfig

logger = logging.getLogger(__name__)


class SlackNotifier(object):
    """Send notifications via slack."""

    def __init__(
        self, api_url: str = "", token: str = "", channel_default: str = ""
    ) -> None:
        """Container for connection with Slack.

        Args:
            api_url (str): url to Slack.
            token (str): token gotten from your app in api.slack.com.
            channel_default (str): name of channel to which you want to post.
        """
        config = JobmonConfig()
        if not api_url:
            api_url = config.get("reaper", "slack_api_url")
        if not token:
            token = config.get("reaper", "slack_token")
        if not channel_default:
            channel_default = config.get("reaper", "slack_channel_default")

        self.api_url = api_url
        self._token = token
        self.channel_default = channel_default

    def send(self, msg: str, channel: Optional[str] = None) -> None:
        """Send message to Slack using requests.post."""
        if channel is None:
            channel = self.channel_default
        resp = requests.post(
            self.api_url,
            headers={"Authorization": "Bearer {}".format(self._token)},
            json={"channel": channel, "text": msg},
        )
        logger.debug(resp)
        if resp.status_code != requests.codes.OK:
            error = "Could not send Slack message. {!r}".format(resp.content)
            # To raise an exception here causes the docker container stop, and
            # becomes hard to restart.
            # Log the error instead. So we can enter the container to fix
            # issues when necessary.
            # Log the status code so that it's easier to identify the cause.
            logger.error(resp.status_code)
            logger.error(error)



"""Routes for Clusters."""
from http import HTTPStatus as StatusCodes
from typing import Any

from flask import jsonify
from sqlalchemy import select

from jobmon.server.web.models.queue import Queue
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint


@blueprint.route("/cluster/<cluster_id>/queue/<queue_name>", methods=["GET"])
def get_queue_by_cluster_queue_names(cluster_id: int, queue_name: str) -> Any:
    """Get the id, name, cluster_name and parameters of a Queue.

    Based on cluster_name and queue_name.
    """
    session = SessionLocal()
    with session.begin():
        select_stmt = select(Queue).where(
            Queue.cluster_id == cluster_id, Queue.name == queue_name
        )
        queue = session.execute(select_stmt).scalars().one_or_none()

    # send back json
    if queue is None:
        resp = jsonify(queue=None)
    else:
        resp = jsonify(queue=queue.to_wire_as_requested_by_client())
    resp.status_code = StatusCodes.OK
    return resp



"""Task Instance Error Log."""
from typing import Tuple

from sqlalchemy import Column, DateTime, ForeignKey, Integer, Text
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func

from jobmon.core.serializers import SerializeTaskInstanceErrorLog
from jobmon.server.web.models import Base


class TaskInstanceErrorLog(Base):
    """The table in the database that logs the error messages for task_instances."""

    __tablename__ = "task_instance_error_log"

    def to_wire(self) -> Tuple:
        """Serialize task instance error log object."""
        return SerializeTaskInstanceErrorLog.to_wire(
            self.id, self.error_time, self.description
        )

    id = Column(Integer, primary_key=True)
    task_instance_id = Column(Integer, ForeignKey("task_instance.id"))
    error_time = Column(DateTime, default=func.now())
    description = Column(Text)

    task_instance = relationship("TaskInstance", back_populates="errors")



"""Routes for Tasks."""
from http import HTTPStatus as StatusCodes
from typing import Any, cast, Dict

from flask import jsonify, request
import sqlalchemy
from sqlalchemy import select
from sqlalchemy.orm import Session
import structlog

from jobmon.core import constants
from jobmon.server.web.models.arg import Arg
from jobmon.server.web.models.task_template import TaskTemplate
from jobmon.server.web.models.task_template_version import TaskTemplateVersion
from jobmon.server.web.models.template_arg_map import TemplateArgMap
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint
from jobmon.server.web.server_side_exception import InvalidUsage


logger = structlog.get_logger(__name__)


@blueprint.route("/task_template", methods=["POST"])
def get_task_template() -> Any:
    """Add a task template for a given tool to the database."""
    # check input variable
    data = cast(Dict, request.get_json())
    try:
        tool_version_id = int(data["tool_version_id"])
        name = data["task_template_name"]
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    structlog.contextvars.bind_contextvars(tool_version_id=tool_version_id)
    logger.info(f"Add task tamplate for tool_version_id {tool_version_id} ")

    # add to DB
    session = SessionLocal()
    try:
        with session.begin():
            task_template = TaskTemplate(tool_version_id=tool_version_id, name=name)
            session.add(task_template)
    except sqlalchemy.exc.IntegrityError:
        with session.begin():
            select_stmt = select(TaskTemplate).where(
                TaskTemplate.tool_version_id == tool_version_id,
                TaskTemplate.name == name,
            )
            task_template = session.execute(select_stmt).scalars().one()

    resp = jsonify(task_template_id=task_template.id)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task_template/<task_template_id>/versions", methods=["GET"])
def get_task_template_versions(task_template_id: int) -> Any:
    """Get the task_template_version."""
    # get task template version object
    structlog.contextvars.bind_contextvars(task_template_id=task_template_id)
    logger.info(f"Getting task template version for task template: {task_template_id}")

    session = SessionLocal()
    with session.begin():
        select_stmt = select(TaskTemplateVersion).where(
            TaskTemplateVersion.task_template_id == task_template_id
        )
        ttvs = session.execute(select_stmt).scalars().all()
        wire_obj = [ttv.to_wire_as_client_task_template_version() for ttv in ttvs]

    resp = jsonify(task_template_versions=wire_obj)
    resp.status_code = StatusCodes.OK
    return resp


def _add_or_get_arg(name: str, session: Session) -> Arg:
    try:
        with session.begin_nested():
            arg = Arg(name=name)
            session.add(arg)
    except sqlalchemy.exc.IntegrityError:
        with session.begin_nested():
            select_stmt = select(Arg).where(Arg.name == name)
            arg = session.execute(select_stmt).scalars().one()
    return arg


@blueprint.route("/task_template/<task_template_id>/add_version", methods=["POST"])
def add_task_template_version(task_template_id: int) -> Any:
    """Add a tool to the database."""
    # check input variables
    structlog.contextvars.bind_contextvars(task_template_id=task_template_id)
    data = cast(Dict, request.get_json())
    try:
        task_template_id = int(task_template_id)
        node_args = data["node_args"]
        task_args = data["task_args"]
        op_args = data["op_args"]
        command_template = data["command_template"].strip()
        arg_mapping_hash = str(data["arg_mapping_hash"]).strip()
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    session = SessionLocal()
    with session.begin():
        # populate the argument table
        arg_mapping_dct: dict = {
            constants.ArgType.NODE_ARG: [],
            constants.ArgType.TASK_ARG: [],
            constants.ArgType.OP_ARG: [],
        }
        for arg_name in node_args:
            arg_mapping_dct[constants.ArgType.NODE_ARG].append(
                _add_or_get_arg(arg_name, session)
            )
        for arg_name in task_args:
            arg_mapping_dct[constants.ArgType.TASK_ARG].append(
                _add_or_get_arg(arg_name, session)
            )
        for arg_name in op_args:
            arg_mapping_dct[constants.ArgType.OP_ARG].append(
                _add_or_get_arg(arg_name, session)
            )

    try:
        with session.begin():
            ttv = TaskTemplateVersion(
                task_template_id=task_template_id,
                command_template=command_template,
                arg_mapping_hash=arg_mapping_hash,
            )
            session.add(ttv)
            session.flush()

            # get a lock
            session.refresh(ttv, with_for_update=True)
            for arg_type_id in arg_mapping_dct.keys():
                for arg in arg_mapping_dct[arg_type_id]:
                    ctatm = TemplateArgMap(
                        task_template_version_id=ttv.id,
                        arg_id=arg.id,
                        arg_type_id=arg_type_id,
                    )
                    session.add(ctatm)
            session.flush()

            task_template_version = ttv.to_wire_as_client_task_template_version()

    except sqlalchemy.exc.IntegrityError:
        with session.begin():
            # if another process is adding this task_template_version then this query should
            # block until the template_arg_map has been populated and committed
            select_stmt = select(TaskTemplateVersion).where(
                TaskTemplateVersion.task_template_id == task_template_id,
                TaskTemplateVersion.command_template == command_template,
                TaskTemplateVersion.arg_mapping_hash == arg_mapping_hash,
            )
            ttv = session.execute(select_stmt).scalars().one()

            task_template_version = ttv.to_wire_as_client_task_template_version()

    resp = jsonify(task_template_version=task_template_version)
    resp.status_code = StatusCodes.OK
    return resp



"""SQLAlchemy database objects."""
from importlib import import_module
from pathlib import Path
from pkgutil import iter_modules
from typing import Any

from sqlalchemy import CheckConstraint, create_engine, event, func, String, text
from sqlalchemy.engine import Engine
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy_utils import database_exists, drop_database
import structlog


logger = structlog.get_logger(__name__)

# declarative registry for model elements
Base: DeclarativeMeta = declarative_base()


@event.listens_for(Base, "instrument_class", propagate=True)
def add_string_length_constraint(Base: DeclarativeMeta, cls_: Any) -> None:
    """Add check constraint to enforce column size limits on sqlite."""
    table = cls_.__table__

    for column in table.columns:
        if isinstance(column.type, String):
            length = column.type.length

            if length is not None:
                logger.debug(
                    f"adding check constraint to {table}.{column} of len={length}"
                )
                CheckConstraint(
                    func.length(column) <= length,
                    table=column,
                )


def load_model() -> None:
    """Iterate through the modules in the current package."""
    package_dir = Path(__file__).resolve().parent
    for _, module_name, _ in iter_modules([str(package_dir)]):
        import_module(f"{__name__}.{module_name}")


def init_db(engine: Engine) -> None:
    """Emit DDL for all modules in 'models'."""
    emit_ddl = True

    # dialect specific init logic
    if engine.dialect.name == "mysql":
        event.remove(Base, "instrument_class", add_string_length_constraint)

        if not engine.url.database:
            raise ValueError("Engine url must include database when calling init_db.")

        if not database_exists(engine.url):
            no_db_url = engine.url._replace(database=None)
            no_db_engine = create_engine(no_db_url)
            query = f"CREATE DATABASE {engine.url.database} CHARACTER SET = 'utf8'"

            with no_db_engine.begin() as connection:
                connection.execute(text(query))
        else:
            emit_ddl = False

    if emit_ddl:
        load_model()

        Base.metadata.create_all(bind=engine)

        # load metadata
        from jobmon.server.web import session_factory
        from jobmon.server.web.models.arg_type import add_arg_types
        from jobmon.server.web.models.cluster_type import add_cluster_types
        from jobmon.server.web.models.cluster import add_clusters
        from jobmon.server.web.models.queue import add_queues
        from jobmon.server.web.models.task_resources_type import (
            add_task_resources_types,
        )
        from jobmon.server.web.models.task_status import add_task_statuses
        from jobmon.server.web.models.task_instance_status import (
            add_task_instance_statuses,
        )
        from jobmon.server.web.models.workflow_status import add_workflow_statuses
        from jobmon.server.web.models.workflow_run_status import (
            add_workflow_run_statuses,
        )

        with session_factory(bind=engine) as session:
            metadata_loaders = [
                add_arg_types,
                add_cluster_types,
                add_clusters,
                add_queues,
                add_task_resources_types,
                add_task_statuses,
                add_task_instance_statuses,
                add_workflow_statuses,
                add_workflow_run_statuses,
            ]
            for loader in metadata_loaders:
                loader(session)
                session.flush()
            session.commit()


def terminate_db(engine: Engine) -> None:
    """Terminate/drop a dev database."""
    # dialect specific init logic

    if database_exists(engine.url):
        drop_database(engine.url)



"""Exceptions only used on the Server Side."""

from typing import Dict, Optional


# Use as base class for server side error
class ServerSideException(Exception):
    """Used for all exceptions on the server side."""

    def __init__(self, msg: str) -> None:
        """Initialize Exception with message."""
        self.msg = msg

    def __str__(self) -> str:
        """Return a printable representational string of the message."""
        return repr(self.msg)


class InvalidUsage(ServerSideException):
    """Error caused by client mistakes (ex. bad data provided)."""

    # TODO: make status_code a parameter for future extension. So far we only use 400
    def __init__(
        self,
        msg: str,
        status_code: Optional[int] = None,
        payload: Optional[tuple] = None,
    ) -> None:
        """Initialize ServerSide exception and default to 400 if not status_code available."""
        super().__init__(msg)
        self.status_code = status_code
        if self.status_code is None:
            # by default, use 400
            self.status_code = 400
        self.payload = payload

    def to_dict(self) -> Dict:
        """Put exception in a dictionary."""
        rv = dict(self.payload or ())
        rv["message"] = self.msg
        return rv


class ServerError(ServerSideException):
    """Use for Internal Server Error."""

    # TODO: make status_code a parameter for future extension. So far we only use 500
    def __init__(
        self,
        msg: str,
        status_code: Optional[int] = None,
        payload: Optional[tuple] = None,
    ) -> None:
        """Initialize and assign 500 status code if none provided."""
        super().__init__(msg)
        self.status_code = status_code
        if self.status_code is None:
            # by default, use 500
            self.status_code = 500
        self.payload = payload

    def to_dict(self) -> Dict:
        """Return payload and message as dictionary."""
        rv = dict(self.payload or ())
        rv["message"] = self.msg
        return rv



"""Cluster Table in the Database."""
from typing import Tuple

from sqlalchemy import Column, ForeignKey, Integer, select, String
from sqlalchemy.orm import relationship, Session

from jobmon.core.serializers import SerializeCluster
from jobmon.server.web.models import Base
from jobmon.server.web.models.cluster_type import ClusterType


class Cluster(Base):
    """Cluster Table in the Database."""

    __tablename__ = "cluster"

    def to_wire_as_requested_by_client(self) -> Tuple:
        """Serialize cluster object."""
        return SerializeCluster.to_wire(
            self.id,
            self.name,
            self.cluster_type.name,
            self.connection_parameters,
        )

    id = Column(Integer, primary_key=True)
    name = Column(String(255), unique=True)
    cluster_type_id = Column(Integer, ForeignKey("cluster_type.id"))
    connection_parameters = Column(String(2500))

    # ORM relationships
    cluster_type = relationship("ClusterType", back_populates="clusters")
    queues = relationship("Queue", back_populates="cluster")


def add_clusters(session: Session) -> None:
    """Populate the cluster table in the database."""
    for cluster_type_name in ["dummy", "sequential", "multiprocess"]:
        cluster_type = (
            session.execute(
                select(ClusterType).where(ClusterType.name == cluster_type_name)
            )
            .scalars()
            .one()
        )
        cluster = Cluster(
            name=cluster_type_name,
            cluster_type_id=cluster_type.id,
            connection_parameters=None,
        )
        session.add(cluster)



"""Task Arg table."""
from sqlalchemy import Column, Integer, String


from jobmon.server.web.models import Base


class TaskArg(Base):
    """Task arg table."""

    __tablename__ = "task_arg"

    task_id = Column(Integer, primary_key=True)
    arg_id = Column(Integer, primary_key=True)
    val = Column(String(2048))



"""Routes for Tool Versions."""
from http import HTTPStatus as StatusCodes
from typing import Any, cast, Dict

from flask import jsonify, request
import sqlalchemy
from sqlalchemy import select
import structlog

from jobmon.server.web.models.task_template import TaskTemplate
from jobmon.server.web.models.tool_version import ToolVersion
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint
from jobmon.server.web.server_side_exception import InvalidUsage


logger = structlog.get_logger(__name__)


@blueprint.route("/tool_version", methods=["POST"])
def add_tool_version() -> Any:
    """Add a new version for a Tool."""
    # check input variable
    data = cast(Dict, request.get_json())

    try:
        tool_id = int(data["tool_id"])
    except Exception as e:
        raise InvalidUsage(
            f"{str(e)} in request to {request.path}", status_code=400
        ) from e

    session = SessionLocal()
    try:
        with session.begin():
            tool_version = ToolVersion(tool_id=tool_id)
            session.add(tool_version)
    except sqlalchemy.exc.IntegrityError:
        with session.begin():
            select_stmt = select(ToolVersion).where(ToolVersion.tool_id == tool_id)
            tool_version = session.execute(select_stmt).scalars().one()

    wire_format = tool_version.to_wire_as_client_tool_version()

    resp = jsonify(tool_version=wire_format)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/tool_version/<tool_version_id>/task_templates", methods=["GET"])
def get_task_templates(tool_version_id: int) -> Any:
    """Get the Tool Version."""
    # check input variable
    structlog.contextvars.bind_contextvars(tool_version_id=tool_version_id)
    logger.info("Getting available task_templates")

    session = SessionLocal()
    with session.begin():
        select_stmt = select(TaskTemplate).where(
            TaskTemplate.tool_version_id == tool_version_id
        )
        task_templates = session.execute(select_stmt).scalars().all()
        wire_format = [t.to_wire_as_client_task_template() for t in task_templates]

    resp = jsonify(task_templates=wire_format)
    resp.status_code = StatusCodes.OK
    return resp



"""Tool DB Table."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship

from jobmon.core.serializers import SerializeClientTool
from jobmon.server.web.models import Base


class Tool(Base):
    """Tool DB Table."""

    __tablename__ = "tool"

    def to_wire_as_client_tool(self) -> tuple:
        """Serialize tool object."""
        serialized = SerializeClientTool.to_wire(id=self.id, name=self.name)
        return serialized

    id = Column(Integer, primary_key=True)
    name = Column(String(255), nullable=False, unique=True)

    tool_versions = relationship("ToolVersion", back_populates="tool")



"""Workflow Attribute Database Table."""
from typing import Any, Dict

from sqlalchemy import Column, ForeignKey, Integer, String

from jobmon.server.web.models import Base


class WorkflowAttribute(Base):
    """Workflow Attribute Database Table."""

    __tablename__ = "workflow_attribute"

    workflow_id = Column(Integer, ForeignKey("workflow.id"), primary_key=True)
    workflow_attribute_type_id = Column(
        Integer, ForeignKey("workflow_attribute_type.id"), primary_key=True
    )
    value = Column(String(255))

    @classmethod
    def from_wire(cls: Any, dct: Dict) -> Any:
        """Create class object from dict."""
        return cls(
            workflow_id=dct["workflow_id"],
            workflow_attribute_type_id=dct["workflow_attribute_type_id"],
            value=dct["value"],
        )

    def to_wire(self) -> Dict:
        """Send data."""
        return {
            "workflow_id": self.workflow_id,
            "workflow_attribute_type_id": self.workflow_attribute_type_id,
            "value": self.value,
        }



"""Template arg map table."""
from sqlalchemy import Column, ForeignKey, Integer
from sqlalchemy.orm import relationship

from jobmon.server.web.models import Base


class TemplateArgMap(Base):
    """Template Arg Map table."""

    __tablename__ = "template_arg_map"

    task_template_version_id = Column(
        Integer, ForeignKey("task_template_version.id"), primary_key=True
    )
    arg_id = Column(Integer, ForeignKey("arg.id"), primary_key=True)
    arg_type_id = Column(Integer, ForeignKey("arg_type.id"), primary_key=True)

    task_template_version = relationship(
        "TaskTemplateVersion", back_populates="template_arg_map"
    )
    argument = relationship("Arg", back_populates="template_arg_map")
    argument_type = relationship("ArgType", back_populates="template_arg_map")



"""Routes for TaskInstances."""
from http import HTTPStatus as StatusCodes
from typing import Any, cast, Dict, Optional

from flask import jsonify, request
import sqlalchemy
from sqlalchemy import select, update
from sqlalchemy.orm import Session
from sqlalchemy.sql import func
import structlog

from jobmon.core import constants
from jobmon.core.exceptions import InvalidStateTransition
from jobmon.core.serializers import SerializeTaskInstanceBatch
from jobmon.server.web._compat import add_time
from jobmon.server.web.models.array import Array
from jobmon.server.web.models.task import Task
from jobmon.server.web.models.task_instance import TaskInstance
from jobmon.server.web.models.task_instance_error_log import TaskInstanceErrorLog
from jobmon.server.web.routes import SessionLocal
from jobmon.server.web.routes.fsm import blueprint
from jobmon.server.web.server_side_exception import ServerError


logger = structlog.get_logger(__name__)


@blueprint.route("/task_instance/<task_instance_id>/log_running", methods=["POST"])
def log_running(task_instance_id: int) -> Any:
    """Log a task_instance as running.

    Args:
        task_instance_id: id of the task_instance to log as running
    """
    structlog.contextvars.bind_contextvars(task_instance_id=task_instance_id)
    data = cast(Dict, request.get_json())

    session = SessionLocal()
    with session.begin():
        select_stmt = select(TaskInstance).where(TaskInstance.id == task_instance_id)
        task_instance = session.execute(select_stmt).scalars().one()

        if data.get("distributor_id", None) is not None:
            task_instance.distributor_id = data["distributor_id"]
        if data.get("nodename", None) is not None:
            task_instance.nodename = data["nodename"]
        task_instance.process_group_id = data["process_group_id"]
        try:
            task_instance.transition(constants.TaskInstanceStatus.RUNNING)
            task_instance.report_by_date = add_time(data["next_report_increment"])
        except InvalidStateTransition as e:
            if task_instance.status == constants.TaskInstanceStatus.RUNNING:
                logger.warning(e)
            elif task_instance.status == constants.TaskInstanceStatus.KILL_SELF:
                task_instance.transition(constants.TaskInstanceStatus.ERROR_FATAL)
            else:
                # Tried to move to an illegal state
                logger.error(e)

        wire_format = task_instance.to_wire_as_worker_node_task_instance()

    resp = jsonify(task_instance=wire_format)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task_instance/<task_instance_id>/log_report_by", methods=["POST"])
def log_ti_report_by(task_instance_id: int) -> Any:
    """Log a task_instance as being responsive with a new report_by_date.

    This is done at the worker node heartbeat_interval rate, so it may not happen at the same
    rate that the reconciler updates batch submitted report_by_dates (also because it causes
    a lot of traffic if all workers are logging report by_dates often compared to if the
    reconciler runs often).

    Args:
        task_instance_id: id of the task_instance to log
    """
    structlog.contextvars.bind_contextvars(task_instance_id=task_instance_id)
    data = cast(Dict, request.get_json())

    session = SessionLocal()
    with session.begin():
        vals = {"report_by_date": add_time(data["next_report_increment"])}
        for optional_val in ["distributor_id", "stderr", "stdout"]:
            val = data.get(optional_val, None)
            if data is not None:
                vals[optional_val] = val

        update_stmt = update(TaskInstance).where(TaskInstance.id == task_instance_id)
        session.execute(update_stmt.values(**vals))
        session.flush()

        select_stmt = select(TaskInstance).where(TaskInstance.id == task_instance_id)
        task_instance = session.execute(select_stmt).scalars().one()
        if task_instance.status == constants.TaskInstanceStatus.TRIAGING:
            task_instance.transition(constants.TaskInstanceStatus.RUNNING)

    resp = jsonify(status=task_instance.status)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task_instance/log_report_by/batch", methods=["POST"])
def log_ti_report_by_batch() -> Any:
    """Log task_instances as being responsive with a new report_by_date.

    This is done at the worker node heartbeat_interval rate, so it may not happen at the same
    rate that the reconciler updates batch submitted report_by_dates (also because it causes
    a lot of traffic if all workers are logging report by_dates often compared to if the
    reconciler runs often).

    Args:
        task_instance_id: id of the task_instance to log
    """
    data = cast(Dict, request.get_json())
    tis = data.get("task_instance_ids", None)

    next_report_increment = float(data["next_report_increment"])

    logger.debug(f"Log report_by for TI {tis}.")
    if tis:
        session = SessionLocal()
        with session.begin():
            update_stmt = (
                update(TaskInstance)
                .where(
                    TaskInstance.id.in_(tis),
                    TaskInstance.status == constants.TaskInstanceStatus.LAUNCHED,
                )
                .values(report_by_date=add_time(next_report_increment))
            )

            session.execute(update_stmt)

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task_instance/<task_instance_id>/log_done", methods=["POST"])
def log_done(task_instance_id: int) -> Any:
    """Log a task_instance as done.

    Args:
        task_instance_id: id of the task_instance to log done
    """
    structlog.contextvars.bind_contextvars(task_instance_id=task_instance_id)
    data = cast(Dict, request.get_json())

    session = SessionLocal()
    with session.begin():
        select_stmt = select(TaskInstance).where(TaskInstance.id == task_instance_id)
        task_instance = session.execute(select_stmt).scalars().one()

        optional_vals = [
            "distributor_id",
            "stdout_log",
            "stderr_log",
            "nodename",
            "stdout",
            "stderr",
        ]
        for optional_val in optional_vals:
            val = data.get(optional_val, None)
            if val is not None:
                setattr(task_instance, optional_val, val)

        try:
            task_instance.transition(constants.TaskInstanceStatus.DONE)
        except InvalidStateTransition as e:
            if task_instance.status == constants.TaskInstanceStatus.DONE:
                logger.warning(e)
            else:
                # Tried to move to an illegal state
                logger.error(e)

    resp = jsonify(status=task_instance.status)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/task_instance/<task_instance_id>/log_error_worker_node", methods=["POST"]
)
def log_error_worker_node(task_instance_id: int) -> Any:
    """Log a task_instance as errored.

    Args:
        task_instance_id (str): id of the task_instance to log done
        error_message (str): message to log as error
    """
    structlog.contextvars.bind_contextvars(task_instance_id=task_instance_id)
    data = cast(Dict, request.get_json())
    logger.info(f"Log ERROR for TI:{task_instance_id}.")

    session = SessionLocal()
    with session.begin():
        select_stmt = select(TaskInstance).where(TaskInstance.id == task_instance_id)
        task_instance = session.execute(select_stmt).scalars().one()

        optional_vals = [
            "distributor_id",
            "stdout_log",
            "stderr_log",
            "nodename",
            "stdout",
            "stderr",
        ]
        for optional_val in optional_vals:
            val = data.get(optional_val, None)
            if data is not None:
                setattr(task_instance, optional_val, val)

        # add error log
        error_state = data["error_state"]
        error_description = data["error_description"]
        try:
            task_instance.transition(error_state)
            error = TaskInstanceErrorLog(
                task_instance_id=task_instance.id, description=error_description
            )
            session.add(error)
        except InvalidStateTransition as e:
            if task_instance.status == error_state:
                logger.warning(e)
            else:
                # Tried to move to an illegal state
                logger.error(e)

    resp = jsonify(status=task_instance.status)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/task_instance/<task_instance_id>/task_instance_error_log", methods=["GET"]
)
def get_task_instance_error_log(task_instance_id: int) -> Any:
    """Route to return all task_instance_error_log entries of the task_instance_id.

    Args:
        task_instance_id (int): ID of the task instance

    Return:
        jsonified task_instance_error_log result set
    """
    structlog.contextvars.bind_contextvars(task_instance_id=task_instance_id)
    logger.info(f"Getting task instance error log for ti {task_instance_id}")

    session = SessionLocal()
    with session.begin():
        select_stmt = (
            select(TaskInstanceErrorLog)
            .where(TaskInstanceErrorLog.task_instance_id == task_instance_id)
            .order_by(TaskInstanceErrorLog.task_instance_id)
        )
        res = session.execute(select_stmt).scalars().all()

    resp = jsonify(task_instance_error_log=[tiel.to_wire() for tiel in res])
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/get_array_task_instance_id/<array_id>/<batch_num>/<step_id>", methods=["GET"]
)
def get_array_task_instance_id(array_id: int, batch_num: int, step_id: int) -> Any:
    """Given an array ID and an index, select a single task instance ID.

    Task instance IDs that are associated with the array are ordered, and selected by index.
    This route will be called once per array task instance worker node, so must be scalable.
    """
    structlog.contextvars.bind_contextvars(array_id=array_id)

    session = SessionLocal()
    with session.begin():
        select_stmt = select(TaskInstance.id).where(
            TaskInstance.array_id == array_id,
            TaskInstance.array_batch_num == batch_num,
            TaskInstance.array_step_id == step_id,
        )
        task_instance_id = session.execute(select_stmt).scalars().one()

    resp = jsonify(task_instance_id=task_instance_id)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/task_instance/<task_instance_id>/log_no_distributor_id", methods=["POST"]
)
def log_no_distributor_id(task_instance_id: int) -> Any:
    """Log a task_instance_id that did not get an distributor_id upon submission."""
    structlog.contextvars.bind_contextvars(task_instance_id=task_instance_id)
    logger.info(
        f"Logging ti {task_instance_id} did not get distributor id upon submission"
    )
    data = cast(Dict, request.get_json())
    logger.debug(f"Log NO DISTRIBUTOR ID. Data {data['no_id_err_msg']}")
    err_msg = data["no_id_err_msg"]

    session = SessionLocal()
    with session.begin():
        select_stmt = select(TaskInstance).where(TaskInstance.id == task_instance_id)
        task_instance = session.execute(select_stmt).scalars().one()
        msg = _update_task_instance_state(
            task_instance, constants.TaskInstanceStatus.NO_DISTRIBUTOR_ID
        )
        error = TaskInstanceErrorLog(
            task_instance_id=task_instance.id, description=err_msg
        )
        session.add(error)

    resp = jsonify(message=msg)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/task_instance/<task_instance_id>/log_distributor_id", methods=["POST"]
)
def log_distributor_id(task_instance_id: int) -> Any:
    """Log a task_instance's distributor id.

    Args:
        task_instance_id: id of the task_instance to log
    """
    structlog.contextvars.bind_contextvars(task_instance_id=task_instance_id)
    data = cast(Dict, request.get_json())
    session = SessionLocal()
    with session.begin():
        select_stmt = select(TaskInstance).where(TaskInstance.id == task_instance_id)
        task_instance = session.execute(select_stmt).scalars().one()
        msg = _update_task_instance_state(
            task_instance, constants.TaskInstanceStatus.LAUNCHED
        )
        task_instance.distributor_id = data["distributor_id"]
        task_instance.report_by_date = add_time(data["next_report_increment"])

    resp = jsonify(message=msg)
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task_instance/<task_instance_id>/log_known_error", methods=["POST"])
def log_known_error(task_instance_id: int) -> Any:
    """Log a task_instance as errored.

    Args:
        task_instance_id (int): id for task instance.
    """
    structlog.contextvars.bind_contextvars(task_instance_id=task_instance_id)
    data = cast(Dict, request.get_json())
    error_state = data["error_state"]
    error_message = data["error_message"]
    distributor_id = data.get("distributor_id", None)
    nodename = data.get("nodename", None)
    logger.info(f"Log ERROR for TI:{task_instance_id}.")

    session = SessionLocal()
    with session.begin():
        select_stmt = select(TaskInstance).where(TaskInstance.id == task_instance_id)
        task_instance = session.execute(select_stmt).scalars().one()

        try:
            resp = _log_error(
                session,
                task_instance,
                error_state,
                error_message,
                distributor_id,
                nodename,
            )
        except sqlalchemy.exc.OperationalError:
            # modify the error message and retry
            new_msg = error_message.encode("latin1", "replace").decode("utf-8")
            resp = _log_error(
                session, task_instance, error_state, new_msg, distributor_id, nodename
            )

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route(
    "/task_instance/<task_instance_id>/log_unknown_error", methods=["POST"]
)
def log_unknown_error(task_instance_id: int) -> Any:
    """Log a task_instance as errored.

    Args:
        task_instance_id (int): id for task instance
    """
    structlog.contextvars.bind_contextvars(task_instance_id=task_instance_id)
    data = cast(Dict, request.get_json())
    error_state = data["error_state"]
    error_message = data["error_message"]
    distributor_id = data.get("distributor_id", None)
    nodename = data.get("nodename", None)
    logger.info(f"Log ERROR for TI:{task_instance_id}.")

    session = SessionLocal()
    with session.begin():
        # make sure the task hasn't logged a new heartbeat since we began
        # reconciliation
        select_stmt = select(TaskInstance).where(
            TaskInstance.id == task_instance_id,
            TaskInstance.report_by_date <= func.now(),
        )
        task_instance = session.execute(select_stmt).scalars().one_or_none()
        session.flush()

        if task_instance is not None:
            try:
                resp = _log_error(
                    session,
                    task_instance,
                    error_state,
                    error_message,
                    distributor_id,
                    nodename,
                )
            except sqlalchemy.exc.OperationalError:
                # modify the error message and retry
                new_msg = error_message.encode("latin1", "replace").decode("utf-8")
                resp = _log_error(
                    session,
                    task_instance,
                    error_state,
                    new_msg,
                    distributor_id,
                    nodename,
                )

    resp = jsonify()
    resp.status_code = StatusCodes.OK
    return resp


@blueprint.route("/task_instance/instantiate_task_instances", methods=["POST"])
def instantiate_task_instances() -> Any:
    """Sync status of given task intance IDs."""
    data = cast(Dict, request.get_json())
    task_instance_ids_list = tuple([int(tid) for tid in data["task_instance_ids"]])

    session = SessionLocal()
    with session.begin():
        # update the task table where FSM allows it
        task_update = (
            update(Task)
            .where(
                Task.id.in_(
                    select(Task.id)
                    .join(TaskInstance, TaskInstance.task_id == Task.id)
                    .where(
                        TaskInstance.id.in_(task_instance_ids_list),
                        (Task.status == constants.TaskStatus.QUEUED),
                    )
                )
            )
            .values(status=constants.TaskStatus.INSTANTIATING, status_date=func.now())
            .execution_options(synchronize_session=False)
        )
        session.execute(task_update)

        # then propagate back into task instance where a change was made
        task_instance_update = (
            update(TaskInstance)
            .where(
                TaskInstance.id.in_(
                    select(TaskInstance.id)
                    .join(Task, TaskInstance.task_id == Task.id)
                    .where(
                        # a successful transition
                        (Task.status == constants.TaskStatus.INSTANTIATING),
                        # and part of the current set
                        TaskInstance.id.in_(task_instance_ids_list),
                    )
                )
            )
            .values(
                status=constants.TaskInstanceStatus.INSTANTIATED, status_date=func.now()
            )
            .execution_options(synchronize_session=False)
        )
        session.execute(task_instance_update)

    with session.begin():
        instantiated_batches_query = (
            select(
                TaskInstance.array_id,
                Array.name,
                TaskInstance.array_batch_num,
                TaskInstance.task_resources_id,
                func.group_concat(TaskInstance.id),
            )
            .where(
                TaskInstance.id.in_(task_instance_ids_list)
                & (TaskInstance.status == constants.TaskInstanceStatus.INSTANTIATED)
                & (TaskInstance.array_id == Array.id)
            )
            .group_by(
                TaskInstance.array_id,
                TaskInstance.array_batch_num,
                TaskInstance.task_resources_id,
                Array.name,
            )
        )
        result = session.execute(instantiated_batches_query)
        serialized_batches = []
        for (
            array_id,
            array_name,
            array_batch_num,
            task_resources_id,
            task_instance_ids,
        ) in result:
            task_instance_ids = [
                int(task_instance_id)
                for task_instance_id in task_instance_ids.split(",")
            ]
            serialized_batches.append(
                SerializeTaskInstanceBatch.to_wire(
                    array_id=array_id,
                    array_name=array_name,
                    array_batch_num=array_batch_num,
                    task_resources_id=task_resources_id,
                    task_instance_ids=task_instance_ids,
                )
            )

    resp = jsonify(task_instance_batches=serialized_batches)
    resp.status_code = StatusCodes.OK
    return resp


# ############################ HELPER FUNCTIONS ###############################
def _update_task_instance_state(task_instance: TaskInstance, status_id: str) -> Any:
    """Advance the states of task_instance and it's associated Task.

    Return any messages that should be published based on the transition.

    Args:
        task_instance (TaskInstance): object of time models.TaskInstance
        status_id (int): id of the status to which to transition
    """
    response = ""
    try:
        task_instance.transition(status_id)
    except InvalidStateTransition:
        if task_instance.status == status_id:
            # It was already in that state, just log it
            msg = (
                f"Attempting to transition to existing state."
                f"Not transitioning task, tid= "
                f"{task_instance.id} from {task_instance.status} to "
                f"{status_id}"
            )
            logger.warning(msg)
            response += msg
        else:
            # Tried to move to an illegal state
            msg = (
                f"Illegal state transition. Not transitioning task, "
                f"tid={task_instance.id}, from {task_instance.status} to "
                f"{status_id}"
            )
            logger.error(msg)
            response += msg
    except Exception as e:
        raise ServerError(
            f"General exception in _update_task_instance_state, jid "
            f"{task_instance}, transitioning to {task_instance}. Not "
            f"transitioning task. Server Error in {request.path}",
            status_code=500,
        ) from e
    return response


def _log_error(
    session: Session,
    ti: TaskInstance,
    error_state: str,
    error_msg: str,
    distributor_id: Optional[int] = None,
    nodename: Optional[str] = None,
) -> Any:
    if nodename is not None:
        ti.nodename = nodename
    if distributor_id is not None:
        ti.distributor_id = distributor_id

    try:
        error = TaskInstanceErrorLog(task_instance_id=ti.id, description=error_msg)
        session.add(error)
        msg = _update_task_instance_state(ti, error_state)
        session.commit()
        resp = jsonify(message=msg)
        resp.status_code = StatusCodes.OK
    except Exception as e:
        session.rollback()
        raise ServerError(
            f"Unexpected Jobmon Server Error in {request.path}", status_code=500
        ) from e

    return resp



"""Task Table for the Database."""

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    Text,
    VARCHAR,
)
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func
import structlog

from jobmon.core.exceptions import InvalidStateTransition
from jobmon.core.serializers import SerializeDistributorTask, SerializeSwarmTask
from jobmon.server.web.models import Base
from jobmon.server.web.models.task_instance_status import TaskInstanceStatus
from jobmon.server.web.models.task_status import TaskStatus


logger = structlog.get_logger(__name__)


class Task(Base):
    """Task Database object."""

    __tablename__ = "task"

    def to_wire_as_distributor_task(self) -> tuple:
        """Serialize executor task object."""
        array_id = self.array.id if self.array is not None else None
        serialized = SerializeDistributorTask.to_wire(
            task_id=self.id,
            array_id=array_id,
            name=self.name,
            command=self.command,
            requested_resources=self.task_resources.requested_resources,
        )
        return serialized

    def to_wire_as_swarm_task(self) -> tuple:
        """Serialize swarm task."""
        serialized = SerializeSwarmTask.to_wire(task_id=self.id, status=self.status)
        return serialized

    id = Column(Integer, primary_key=True)
    workflow_id = Column(Integer, ForeignKey("workflow.id"))
    node_id = Column(Integer, ForeignKey("node.id"))
    task_args_hash = Column(VARCHAR(50), index=True)
    array_id = Column(Integer, ForeignKey("array.id"), default=None)
    name = Column(String(255), index=True)
    command = Column(Text)
    task_resources_id = Column(Integer, ForeignKey("task_resources.id"), default=None)
    num_attempts = Column(Integer, default=0)
    max_attempts = Column(Integer, default=1)
    resource_scales = Column(String(1000), default=None)
    fallback_queues = Column(String(1000), default=None)
    status = Column(String(1), ForeignKey("task_status.id"))
    status_date = Column(DateTime, default=func.now(), index=True)

    # ORM relationships
    task_instances = relationship("TaskInstance", back_populates="task")
    task_resources = relationship("TaskResources", foreign_keys=[task_resources_id])
    array = relationship("Array", foreign_keys=[array_id])

    __table_args__ = (
        Index("ix_workflow_id_status_date", "workflow_id", "status_date"),
    )

    # Finite state machine
    valid_transitions = [
        (TaskStatus.REGISTERING, TaskStatus.QUEUED),
        (TaskStatus.ADJUSTING_RESOURCES, TaskStatus.QUEUED),
        (TaskStatus.QUEUED, TaskStatus.INSTANTIATING),
        (TaskStatus.INSTANTIATING, TaskStatus.LAUNCHED),
        (TaskStatus.INSTANTIATING, TaskStatus.ERROR_RECOVERABLE),
        (TaskStatus.LAUNCHED, TaskStatus.RUNNING),
        (TaskStatus.LAUNCHED, TaskStatus.ERROR_RECOVERABLE),
        (TaskStatus.INSTANTIATING, TaskStatus.ERROR_RECOVERABLE),
        (TaskStatus.INSTANTIATING, TaskStatus.RUNNING),
        (TaskStatus.RUNNING, TaskStatus.DONE),
        (TaskStatus.RUNNING, TaskStatus.ERROR_RECOVERABLE),
        (TaskStatus.ERROR_RECOVERABLE, TaskStatus.ADJUSTING_RESOURCES),
        (TaskStatus.ERROR_RECOVERABLE, TaskStatus.QUEUED),
        (TaskStatus.ERROR_RECOVERABLE, TaskStatus.ERROR_FATAL),
        (TaskStatus.ERROR_RECOVERABLE, TaskStatus.REGISTERING),
    ]

    def reset(
        self, name: str, command: str, max_attempts: int, reset_if_running: bool
    ) -> None:
        """Reset status and number of attempts on a Task."""
        # only reset undone tasks
        if self.status != TaskStatus.DONE:
            # only reset if the task is not currently running or if we are
            # resetting running tasks
            if self.status != TaskStatus.RUNNING or reset_if_running:
                self.status = TaskStatus.REGISTERING
                self.num_attempts = 0
                self.name = name
                self.command = command
                self.max_attempts = max_attempts
                self.status_date = func.now()

    def transition(self, new_state: str) -> None:
        """Transition the Task to a new state."""
        # bind_to_logger(workflow_id=self.workflow_id, task_id=self.id)
        logger.info(f"Transitioning task from {self.status} to {new_state}")
        self._validate_transition(new_state)
        if new_state == TaskStatus.QUEUED:
            self.num_attempts = self.num_attempts + 1
        self.status = new_state
        self.status_date = func.now()

    def transition_after_task_instance_error(
        self, job_instance_error_state: str
    ) -> None:
        """Transition the task to an error state."""
        # bind_to_logger(workflow_id=self.workflow_id, task_id=self.id)
        logger.info("Transitioning task to ERROR_RECOVERABLE")
        self.transition(TaskStatus.ERROR_RECOVERABLE)
        if self.num_attempts >= self.max_attempts:
            logger.info("Giving up task after max attempts.")
            self.transition(TaskStatus.ERROR_FATAL)
        else:
            if job_instance_error_state == TaskInstanceStatus.RESOURCE_ERROR:
                logger.debug("Adjust resource for task.")
                self.transition(TaskStatus.ADJUSTING_RESOURCES)
            else:
                logger.debug("Retrying Task.")
                self.transition(TaskStatus.REGISTERING)

    def _validate_transition(self, new_state: str) -> None:
        """Ensure the task state transition is valid."""
        if (self.status, new_state) not in self.valid_transitions:
            raise InvalidStateTransition("Task", self.id, self.status, new_state)



"""Node Table in the Database."""
from sqlalchemy import Column, Integer, UniqueConstraint, VARCHAR

from jobmon.server.web.models import Base


class Node(Base):
    """Node Table in the Database."""

    __tablename__ = "node"

    id = Column(Integer, primary_key=True)
    task_template_version_id = Column(Integer, index=True)
    node_args_hash = Column(VARCHAR(50))

    __table_args__ = (
        UniqueConstraint(
            "task_template_version_id",
            "node_args_hash",
            name="uc_task_template_version_id_node_args_hash",
        ),
    )



"""Routes used to move through the finite state."""
from typing import Optional

from flask import Blueprint

from jobmon.server.web import routes
from jobmon.server.web.routes import SessionLocal

blueprint = Blueprint("finite_state_machine", __name__)
blueprint.add_url_rule("/", view_func=routes.is_alive, methods=["GET"])
blueprint.add_url_rule("/time", view_func=routes.get_pst_now, methods=["GET"])
blueprint.add_url_rule("/health", view_func=routes.health, methods=["GET"])
blueprint.add_url_rule(
    "/test_bad", view_func=routes.test_route, methods=["GET"]  # type: ignore
)

from jobmon.server.web.routes.fsm import (
    array,
    cluster,
    dag,
    node,
    queue,
    task,
    task_instance,
    task_resources,
    task_template,
    tool,
    tool_version,
    workflow,
    workflow_run,
)


@blueprint.teardown_request
def teardown(e: Optional[BaseException]) -> None:
    """Remove threadlocal session from registry."""
    SessionLocal.remove()



"""Reaper Behavior for a given Workflow Run."""
from __future__ import annotations

import logging
from typing import Any

from jobmon.core.exceptions import InvalidResponse
from jobmon.core.requester import http_request_ok, Requester
from jobmon.core.serializers import SerializeWorkflowRun


logger = logging.getLogger(__file__)


class ReaperWorkflowRun(object):
    """Reaper Behavior for a given Workflow Run."""

    def __init__(
        self, workflow_run_id: int, workflow_id: int, requester: Requester
    ) -> None:
        """Implementing workflow reaper behavior of workflow run.

        Args:
            workflow_run_id (int): id of workflow run object from DB.
            workflow_id (int): id of associated workflow.
            requester (Requester): requester to communicate with Flask.
        """
        self.workflow_run_id = workflow_run_id
        self.workflow_id = workflow_id
        self._requester = requester

    @classmethod
    def from_wire(
        cls: Any, wire_tuple: tuple, requester: Requester
    ) -> ReaperWorkflowRun:
        """Create Reaper Workflow Run object."""
        kwargs = SerializeWorkflowRun.kwargs_from_wire(wire_tuple)
        return cls(
            workflow_run_id=kwargs["id"],
            workflow_id=kwargs["workflow_id"],
            requester=requester,
        )

    def reap(self) -> str:
        """Transition workflow run to error."""
        app_route = f"/workflow_run/{self.workflow_run_id}/reap"
        return_code, response = self._requester.send_request(
            app_route=app_route, message={}, request_type="put"
        )
        if http_request_ok(return_code) is False:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from PUT "
                f"request through route {app_route}. Expected "
                f"code 200. Response content: {response}"
            )
        return response["status"]

    def __repr__(self) -> str:
        """Return formatted reaper workflow run data."""
        return (
            f"ReaperWorkflowRun(workflow_run_id={self.workflow_run_id}, "
            f"workflow_id={self.workflow_id}"
        )
